package ahc.collection;

import ahc.collection.access.AccessStrategy;
import ahc.collection.store.Store;


/**
 * This is the generic, configurable implementation of both List and Set.
 *
 * @author Arno Haase
 */
public class GenericCollection implements List, Set {
    private final Store _store;
    private final AccessStrategy _accessStrategy;

    public GenericCollection (Store store, AccessStrategy accessStrategy) {
        _store = store;
        _accessStrategy = accessStrategy;
    }

    // general info methods

    public int size () {
        return _store.size ();
    }

    public boolean isEmpty () {
        return _store.isEmpty ();
    }

    AccessStrategy getAccessStrategy () {
        return _accessStrategy;
    }

    Store getStore () {
        return _store;
    }


    // read access methods

    public boolean contains (Object o) {
        return _accessStrategy.first (_store, o) != null;
    }

    public boolean containsAll (java.util.Collection c) {
        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();) {
            if (!contains (iter.next ()))
                return false;
        }
        return true;
    }

    public Object[] toArray () {
        return _store.toArray ();
    }

    public Object[] toArray (Object a[]) {
        return _store.toArray (a);
    }

    public Object get (int index) {
        return _accessStrategy.get (_store, index);
    }

    public int indexOf (Object o) {
        final Iterator iter = _accessStrategy.first (_store, o);
        return (iter != null) ? iter.currentIndex () : -1;
    }

    public int firstIndexOf (Object o) {
        return indexOf (o);
    }

    public int lastIndexOf (Object o) {
        final Iterator iter = _accessStrategy.last (_store, o);
        return (iter != null) ? iter.currentIndex () : -1;
    }


    // modifying methods

    public boolean add (Object o) {
        return _accessStrategy.add (_store, o);
    }

    public boolean addAll (java.util.Collection c) {
        boolean result = false;

        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();) {
            if (_accessStrategy.add (_store, iter.next ()))
                result = true;
        }
        return result;
    }

    public void add (int index, Object element) {
        _accessStrategy.add (_store, _store.iter (index), element);
    }

    public boolean addAll (int index, java.util.Collection c) {
        boolean result = false;
        final Iterator dest = _store.iter (index);
        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();) {
            if (_accessStrategy.add (_store, dest, iter.next ()))
                result = true;
        }
        return result;
    }

    public boolean prepend (Object o) {
        return _accessStrategy.prepend (_store, o);
    }

    public boolean append (Object o) {
        return _accessStrategy.append (_store, o);
    }


    public Object set (int index, Object element) {
        final Iterator iter = _store.iter (index);
        final Object result = iter.current ();
        iter.set (element);
        return result;
    }


    public boolean remove (Object o) {
        return _store.remove (o);
    }


    public Object remove (int index) {
        final Iterator iter = _store.iter (index);
        final Object result = iter.current ();
        iter.remove ();
        return result;
    }

    public boolean removeAll (java.util.Collection c) {
        boolean result = false;
        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();) {
            if (_accessStrategy.remove (_store, iter.next ()))
                result = true;
        }
        return result;
    }

    public boolean retainAll (java.util.Collection c) {
        final int oldSize = size ();
        for (Iterator iter = begin (); iter.hasCurrent (); iter.next ()) {
            if (!c.contains (iter.current ())) {
                iter.remove ();
            }
        }
        return oldSize != size ();
    }

    public void clear () {
        _store.clear ();
    }


    // iterator stuff

    public Iterator before () {
        return _accessStrategy.before (_store);
    }

    public Iterator begin () {
        return _accessStrategy.begin (_store);
    }

    public Iterator end () {
        return _accessStrategy.end (_store);
    }

    public Iterator after () {
        return _accessStrategy.after (_store);
    }

    public Iterator iter (int index) {
        return _accessStrategy.iter (_store, index);
    }

    public Iterator first (Object o) {
        return _accessStrategy.first (_store, o);
    }

    public Iterator last (Object o) {
        return _accessStrategy.last (_store, o);
    }

    public java.util.Iterator iterator () {
        return before ();
    }

    public java.util.ListIterator listIterator () {
        return before ();
    }

    public java.util.ListIterator listIterator (int index) {
        return iter (index);
    }


    //*****todo ???
    public java.util.List subList (int fromIndex, int toIndex) {
        throw new UnsupportedOperationException ();
    }

    public String toString () {
        final StringBuffer result = new StringBuffer ("[");

        boolean first = true;
        for (Iterator iter = begin(); iter.hasCurrent(); iter.next()) {
            if (!first)
                result.append (", ");
            first = false;

            result.append (iter.current());
        }

        result.append ("]");
        return result.toString();
    }
}
