package ahc.collection;

import ahc.collection.access.AccessStrategy;
import ahc.collection.equality.EqualityStrategy;
import ahc.collection.store.Store;


/**
 * This is the generic, configurable implementation of a Map.
 *
 * @author Arno Haase
 */
public class GenericMap implements Map {
    private final Store _store;
    private final AccessStrategy _accessStrat;
    private final EqualityStrategy _keyEqualityStrategy;
    private final EqualityStrategy _valueEqualityStrategy;


    public GenericMap (Store store, AccessStrategy accessStrat) {
        this (store, accessStrat, store.getEqualityStrategy ());
    }

    public GenericMap (Store store, AccessStrategy accessStrat, EqualityStrategy valueEqualityStrategy) {
        _store = store;
        _accessStrat = accessStrat;
        _keyEqualityStrategy = store.getEqualityStrategy ();
        _valueEqualityStrategy = valueEqualityStrategy;

        _store.setEqualityStrategy (new EntryEqualityStrategy (store.getEqualityStrategy ()));
    }


    public EqualityStrategy getValueEqualityStrategy () {
        return _valueEqualityStrategy;
    }


    public int size () {
        return _store.size ();
    }

    public boolean isEmpty () {
        return _store.isEmpty ();
    }

    public boolean containsKey (Object key) {
        return findKey (key) != null;
    }

    public boolean containsValue (Object value) {
        for (Iterator iter = _accessStrat.begin (_store); iter.hasCurrent (); iter.next ()) {
            if (_valueEqualityStrategy.equals (value, ((EntryImpl) iter.current ())._value))
                return true;
        }
        return false;
    }

    public Object get (Object key) {
        final Iterator iter = findKey (key);
        return (iter == null) ? null : ((EntryImpl) iter.current ())._value;
    }

    public Object put (Object key, Object value) {
        final EntryImpl newEntry = new EntryImpl (key, value);

        final Iterator old = _accessStrat.first (_store, newEntry);
        final Object result = (old == null) ? null : ((Map.Entry) old.current ()).getValue ();

        if (!_accessStrat.add (_store, newEntry)) {
            old.remove ();
            _accessStrat.add (_store, newEntry);
        }

        return result;
    }

    protected Iterator findKey (Object key) {
        return _accessStrat.first (_store, new EntryImpl (key, null));
    }

    public Object remove (Object key) {
        Object result = null;
        Iterator iter = null;

        while ((iter = findKey (key)) != null) {
            result = ((EntryImpl) iter.current ())._value;
            iter.remove ();
        }

        return result;
    }

    public void putAll (java.util.Map map) {
        for (java.util.Iterator iter = map.entrySet ().iterator (); iter.hasNext ();) {
            final Map.Entry current = (Map.Entry) iter.next ();
            put (current.getKey (), current.getValue ());
        }
    }

    public void clear () {
        _store.clear ();
    }

    public Set keys () {
        return new KeySet ();
    }

    public java.util.Set keySet () {
        return keys ();
    }


    public Collection vals () {
        return new ValueCollection ();
    }

    public java.util.Collection values () {
        return vals ();
    }


    public Set entries () {
        return new EntrySet ();
    }

    public java.util.Set entrySet () {
        return entries ();
    }


    public String toString () {
        final StringBuffer result = new StringBuffer ("[");

        boolean first = true;
        for (Iterator iter = entries ().begin (); iter.hasCurrent (); iter.next ()) {
            if (!first)
                result.append (", ");
            first = false;

            result.append (((Entry) iter.current ()).getKey () + "->" + ((Entry) iter.current ()).getValue ());
        }

        result.append ("]");
        return result.toString ();
    }


    protected class EntrySet extends UnmodifyableSet {
        public Iterator before () {
            return new EntrySetIterator (_accessStrat.before (_store));
        }

        public boolean contains (Object o) {
            if (!(o instanceof Map.Entry))
                return false;
            final Entry entry = (Entry) o;

            for (Iterator iter = begin (); iter.hasCurrent (); iter.next ()) {
                final Entry curEntry = (Map.Entry) iter.current ();
                if (_keyEqualityStrategy.equals (entry.getKey (), curEntry.getKey ()) &&
                        _valueEqualityStrategy.equals (entry.getValue (), curEntry.getValue ()))
                    return true;
            }
            return false;
        }
    }


    protected class EntrySetIterator extends UnmodifyableIterator {
        public EntrySetIterator (Iterator inner) {
            super (inner);
        }

        public Object current () {
            return _inner.current ();
        }

        public Object next () {
            return _inner.next ();
        }

        public Object next (int n) {
            return _inner.next (n);
        }
    }


    protected class ValueCollection extends UnmodifyableSet {
        public Iterator before () {
            return new ValueCollectionIterator (_accessStrat.before (_store));
        }

        public boolean contains (Object o) {
            for (Iterator iter = begin (); iter.hasCurrent (); iter.next ()) {
                if (_valueEqualityStrategy.equals (iter.current(), o))
                    return true;
            }
            return false;
        }
    }


    protected class ValueCollectionIterator extends UnmodifyableIterator {
        public ValueCollectionIterator (Iterator inner) {
            super (inner);
        }

        public Object current () {
            return safeValue ((Map.Entry) _inner.current ());
        }

        public Object next () {
            return safeValue ((Map.Entry) _inner.next ());
        }

        public Object next (int n) {
            return safeValue ((Map.Entry) _inner.next (n));
        }
    }


    protected class KeySet extends UnmodifyableSet {
        public Iterator before () {
            return new KeySetIterator (_accessStrat.before (_store));
        }

        public boolean contains (Object o) {
            return _accessStrat.first (_store, new EntryImpl (o, null)) != null;
        }
    }


    protected class KeySetIterator extends UnmodifyableIterator {
        public KeySetIterator (Iterator inner) {
            super (inner);
        }

        public Object current () {
            return safeKey ((Map.Entry) _inner.current ());
        }

        public Object next () {
            return safeKey ((Map.Entry) _inner.next ());
        }

        public Object next (int n) {
            return safeKey ((Map.Entry) _inner.next (n));
        }
    };


    protected abstract class UnmodifyableIterator implements Iterator {
        protected final Iterator _inner;

        Object safeKey (Map.Entry entry) {
            if (entry == null)
                return null;

            return entry.getKey ();
        }

        Object safeValue (Map.Entry entry) {
            if (entry == null)
                return null;

            return entry.getValue ();
        }

        public UnmodifyableIterator (Iterator inner) {
            _inner = inner;
        }

        public boolean hasCurrent () {
            return _inner.hasCurrent ();
        }

        public int currentIndex () {
            return _inner.currentIndex ();
        }

        public boolean replace (Object o) {
            throw new UnsupportedOperationException ();
        }

        public boolean hasNext () {
            return _inner.hasNext ();
        }

        public boolean hasPrevious () {
            return _inner.hasPrevious ();
        }

        public int nextIndex () {
            return _inner.nextIndex ();
        }

        public Object previous (int n) {
            throw new UnsupportedOperationException ();
        }

        public Object previous () {
            throw new UnsupportedOperationException ();
        }

        public int previousIndex () {
            return _inner.previousIndex ();
        }

        public void remove () {
            throw new UnsupportedOperationException ();
        }

        public void set (Object o) {
            throw new UnsupportedOperationException ();
        }

        public void add (Object o) {
            throw new UnsupportedOperationException ();
        }
    }


    protected abstract class UnmodifyableSet implements Set {

        public Iterator begin () {
            final Iterator iter = before ();
            iter.next ();
            return iter;
        }

        public int size () {
            return _store.size ();
        }

        public boolean isEmpty () {
            return _store.isEmpty ();
        }

        public java.util.Iterator iterator () {
            return before ();
        }

        public Object[] toArray () {
            return toArray (new Object[size ()]);
        }

        public Object[] toArray (Object a[]) {
            if (a.length < size ())
                a = (Object[]) java.lang.reflect.Array.newInstance (a.getClass ().getComponentType (), size ());

            int ind = 0;
            for (Iterator iter = before (); iter.hasNext ();) {
                a[ind++] = iter.next ();
            }
            return a;
        }

        public String toString () {
            return new java.util.HashSet (this).toString ();
        }

        public boolean add (Object o) {
            throw new UnsupportedOperationException ();
        }

        public boolean remove (Object o) {
            throw new UnsupportedOperationException ();
        }

        public boolean containsAll (java.util.Collection c) {
            throw new UnsupportedOperationException ();
        }

        public boolean addAll (java.util.Collection c) {
            throw new UnsupportedOperationException ();
        }

        public boolean removeAll (java.util.Collection c) {
            throw new UnsupportedOperationException ();
        }

        public boolean retainAll (java.util.Collection c) {
            throw new UnsupportedOperationException ();
        }

        public void clear () {
            throw new UnsupportedOperationException ();
        }
    }


    protected class EntryEqualityStrategy implements EqualityStrategy {
        private final EqualityStrategy _inner;

        public EntryEqualityStrategy (EqualityStrategy inner) {
            _inner = inner;
        }

        public boolean equals (Object o1, Object o2) {
            final Map.Entry e1 = (Map.Entry) o1;
            final Map.Entry e2 = (Map.Entry) o2;

            return _inner.equals (e1.getKey (), e2.getKey ());
        }

        public int hashCode (Object o) {
            final Map.Entry e = (Map.Entry) o;
            return _inner.hashCode (e.getKey ());
        }
    }


    protected class EntryImpl implements Map.Entry {
        private final Object _key;
        private Object _value;

        public EntryImpl (Object key, Object value) {
            _key = key;
            _value = value;
        }

        public Object getKey () {
            return _key;
        }

        public Object getValue () {
            return _value;
        }

        public Object setValue (Object value) {
            Object old = _value;
            _value = value;
            return old;
        }

        public String toString () {
            return "" + _key + "->" + _value;
        }

        public int hashCode () {
            return _keyEqualityStrategy.hashCode (_key) ^
                    _valueEqualityStrategy.hashCode (_value);
        }

        public boolean equals (Object o) {
            if (o == null || !(o instanceof Entry))
                return false;
            final Entry other = (Entry) o;

            return _keyEqualityStrategy.equals (_key, other.getKey ()) &&
                    _valueEqualityStrategy.equals (_value, other.getValue ());
        }
    }
}
