package ahc.collection;

import ahc.util.IntegerFactory;

import java.util.Collection;


/**
 * This is the generic, configurable implementation of a MultiSet.
 *
 * @author Arno Haase
 */
public class GenericMultiSet implements MultiSet {
    private final Map _map;


    public GenericMultiSet (Map map) {
        if (! map.isEmpty())
            throw new IllegalArgumentException ("map must be empty");
        _map = map;
    }


    public int getCount (Object o) {
        Integer count = getInt (o);
        return (count == null) ? 0 : count.intValue ();
    }

    private Integer getInt (Object o) {
        final Integer count = (Integer) _map.get (o);
        return count;
    }

    public boolean removeSingle (Object o) {
        final Integer count = getInt (o);
        if (count == null)
            return false;
        if (count.intValue () == 1) {
            _map.remove (o);
            return true;
        }

        _map.put (o, IntegerFactory.decrement (count));
        return true;
    }

    public boolean add (Object o) {
        final Integer count = getInt (o);
        _map.put (o, IntegerFactory.increment (count));
        return true;
    }


    public Iterator before () {
        return new MultiSetIterator (_map.keys ().before ());
    }

    public Iterator begin () {
        final Iterator result = before ();
        result.next ();
        return result;
    }


    protected class MultiSetIterator implements Iterator {
        private final Iterator _keysIter;

        public MultiSetIterator (Iterator keysIter) {
            _keysIter = keysIter;
        }

        public boolean hasCurrent () {
            return _keysIter.hasCurrent ();
        }

        public Object current () {
            return _keysIter.current ();
        }


        public Object next (int n) {
            _keysIter.next (n);
            return safeCurrent ();
        }


        public boolean hasNext () {
            return _keysIter.hasNext ();
        }

        public Object next () {
            _keysIter.next ();
            return safeCurrent ();
        }

        private Object safeCurrent () {
            return (_keysIter.hasCurrent ()) ? _map.get (_keysIter.current ()) : null;
        }

        public int currentIndex () {
            throw new UnsupportedOperationException ();
        }

        public Object previous (int n) {
            throw new UnsupportedOperationException ();
        }

        public boolean replace (Object o) {
            throw new UnsupportedOperationException ();
        }

        public boolean hasPrevious () {
            throw new UnsupportedOperationException ();
        }

        public Object previous () {
            throw new UnsupportedOperationException ();
        }

        public int nextIndex () {
            throw new UnsupportedOperationException ();
        }

        public int previousIndex () {
            throw new UnsupportedOperationException ();
        }

        public void remove () {
            throw new UnsupportedOperationException ();
        }

        public void set (Object o) {
            throw new UnsupportedOperationException ();
        }

        public void add (Object o) {
            throw new UnsupportedOperationException ();
        }
    }


    public int size () {
        return _map.size ();
    }

    public boolean isEmpty () {
        return _map.isEmpty ();
    }

    public boolean contains (Object o) {
        return _map.containsKey (o);
    }

    public java.util.Iterator iterator () {
        return before ();
    }

    public Object[] toArray () {
        return _map.keys ().toArray ();
    }

    public Object[] toArray (Object a[]) {
        return _map.keys ().toArray (a);
    }

    public boolean remove (Object o) {
        return _map.remove (o) != null;
    }

    public boolean containsAll (java.util.Collection c) {
        return _map.keys ().containsAll (c);
    }

    public boolean addAll (java.util.Collection c) {
        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();)
            add (iter.next ());
        return true;
    }

    public boolean removeAll (java.util.Collection c) {
        boolean modified = false;
        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();) {
            if (remove (iter.next ()))
                modified = true;
        }
        return modified;
    }

    public boolean retainAll (java.util.Collection c) {
        boolean modified = false;
        for (Iterator iter = begin (); iter.hasCurrent (); iter.next ()) {
            if (!c.contains (iter.current ())) {
                remove (iter.current ());
                modified = true;
            }
        }
        return modified;
    }

    public void clear () {
        _map.clear ();
    }


    public void addAll (MultiSet m) {
        for (Iterator iter = m.begin (); iter.hasCurrent (); iter.next ()) {
            final Integer curValue = getInt (iter.current ());
            _map.put (iter.current (), IntegerFactory.add (curValue, m.getCount (iter.current ())));
        }
    }

    public void removeAll (MultiSet m) {
        for (Iterator iter = m.begin (); iter.hasCurrent (); iter.next ()) {
            Integer curValue = getInt (iter.current ());
            curValue = IntegerFactory.subtract (curValue, m.getCount (iter.current ()));
            if (curValue.intValue () < 0)
                curValue = IntegerFactory.get (0);
            _map.put (iter.current (), curValue);
        }
    }

    public boolean removeAllSingle (Collection c) {
        boolean modified = false;
        for (java.util.Iterator iter = c.iterator (); iter.hasNext ();) {
            if (removeSingle (iter.next ()))
                modified = true;
        }
        return modified;
    }
}
