package ahc.collection;

import ahc.collection.store.Store;
import ahc.collection.access.AccessStrategy;
import ahc.collection.access.GenericAccessStrategy;
import ahc.collection.Pair;
import ahc.util.SynchronizedParamGetCode;
import ahc.util.NaturalOrderComparator;
import ahc.exception.TimeoutException;

import java.util.Comparator;
import java.util.NoSuchElementException;


/**
 * This is the generic, configurable implementation of a PriorityQueue.
 *
 * @author Arno Haase
 */
public class GenericPriorityQueue implements PriorityQueue {
    private final Store _store;
    private final boolean _isThreadSafe;
    private final Comparator _comparator;
    private final AccessStrategy _accessStrategy;

    private Object _threshold = null;

    final SynchronizedParamGetCode _peekCode;
    final SynchronizedParamGetCode _removeCode;


    //todo: natural ordering in factory
    //todo: externally configurable AccessStrategy


    public GenericPriorityQueue (Store store, boolean threadSafe) {
        this (store, NaturalOrderComparator.INSTANCE, threadSafe);
    }

    public GenericPriorityQueue (Store store, Comparator comparator, boolean threadSafe) {
        if (!store.preservesOrdering () && !store.isSorted ()) {
            throw new IllegalArgumentException ("the store for a PriorityQueue must be either sorted or preserve order");
        }

        if (!store.isEmpty ())
            throw new IllegalArgumentException ("store must be empty on collection creation");

        _store = store;
        _isThreadSafe = threadSafe;
        _comparator = comparator;
        _accessStrategy = new GenericAccessStrategy (new PriorityComparator (comparator), false);

        _removeCode = new SynchronizedParamGetCode (_store) {
            public boolean precondition (Object threshold) {
                return !isEmpty (threshold);
            }

            public Object getUnsynchronized (Object minPriority) {
                final Iterator iter = _accessStrategy.end (_store);
                if (!iter.hasCurrent ())
                    throw new NoSuchElementException ();

                final Pair result = (Pair) iter.current ();
                if (!passesThreshold (result.getFirst (), minPriority))
                    throw new NoSuchElementException ();

                iter.remove ();
                return result.getSecond ();
            }
        };
        _peekCode = new SynchronizedParamGetCode (_store) {
            public boolean precondition (Object threshold) {
                return !isEmpty (threshold);
            }

            public Object getUnsynchronized (Object minPriority) {
                if (_store.isEmpty ())
                    throw new NoSuchElementException ();

                final Iterator iter = _accessStrategy.end (_store);
                final Pair result = (Pair) iter.current ();
                if (!passesThreshold (result.getFirst (), minPriority))
                    throw new NoSuchElementException ();

                return result.getSecond ();
            }
        };
    }


    public boolean add (Object obj, Object priority) {
        if (_isThreadSafe) {
            synchronized (_store) {
                if (!passesThreshold (priority))
                    return false;

                _store.notify ();
                return _accessStrategy.prepend (_store, new Pair (priority, obj));
            }
        }
        else {
            if (!passesThreshold (priority))
                return false;

            return _accessStrategy.prepend (_store, new Pair (priority, obj));
        }
    }

    private boolean passesThreshold (Object priority) {
        if (priority == null)
            throw new IllegalArgumentException ("priority must not be null");

        return passesThreshold (priority, _threshold);
    }

    private boolean passesThreshold (Object priority, Object threshold) {
        if (threshold == null)
            return true;

        return _comparator.compare (priority, threshold) >= 0;
    }


    public Object remove () {
        return remove (null);
    }

    public Object remove (boolean blocking) throws InterruptedException {
        return remove (null, blocking);
    }

    public Object remove (int timeoutMillis) throws InterruptedException, TimeoutException {
        return remove (null, timeoutMillis);
    }

    public Object remove (Object minPriority) {
        if (_isThreadSafe)
            throw new IllegalStateException ("for threadsafe queue use remove (boolean) instead.");

        return _removeCode.getUnsynchronized (minPriority);
    }

    public Object remove (Object minPriority, boolean blocking) throws InterruptedException {
        if (isThreadSafe ()) {
            return _removeCode.get (minPriority, blocking);
        }
        else
            throw new IllegalStateException ("makes sense only for threadsafe queues");
    }


    public Object remove (Object minPriority, int timeoutMillis) throws InterruptedException, TimeoutException {
        if (isThreadSafe ()) {
            return _removeCode.get (minPriority, timeoutMillis);
        }
        else
            throw new IllegalStateException ("makes sense only for threadsafe queues");
    }

    public Object peek () {
        return peek (null);
    }

    public Object peek (boolean blocking) throws InterruptedException {
        return peek (null, blocking);
    }

    public Object peek (int timeoutMillis) throws InterruptedException, TimeoutException {
        return peek (null, timeoutMillis);
    }

    public Object peek (Object minPriority) {
        if (_isThreadSafe)
            throw new IllegalStateException ("for threadsafe queue use peek (boolean) instead.");

        return _peekCode.getUnsynchronized(minPriority);
    }


    public Object peek (Object minPriority, boolean blocking) throws InterruptedException {
        if (isThreadSafe ()) {
                return _peekCode.get (minPriority, blocking);
        }
        else
            throw new IllegalStateException ("makes sense only for threadsafe queues");
    }

    public Object peek (Object minPriority, int timeoutMillis) throws InterruptedException, TimeoutException {
        if (isThreadSafe ()) {
            return _peekCode.get (minPriority, timeoutMillis);
        }
        else
            throw new IllegalStateException ("makes sense only for threadsafe queues");
    }

    public int size () {
        return _store.size ();
    }

    public int size (Object minPriority) {
        int numBelow = 0;
        final Iterator iter = _accessStrategy.begin (_store);
        while (iter.hasCurrent () && !passesThreshold (((Pair) iter.current ()).getFirst (), minPriority)) {
            iter.next ();
            numBelow++;
        }

        return size () - numBelow;
    }

    public boolean isEmpty () {
        return _store.isEmpty ();
    }

    public boolean isEmpty (Object minPriority) {
        //todo: optimize lookup to log(n) --> AccessStrategy.firstGreaterEquals
        final Iterator iter = _accessStrategy.begin (_store);
        while (iter.hasCurrent () && !passesThreshold (((Pair) iter.current ()).getFirst (), minPriority))
            iter.next ();

        return !iter.hasCurrent ();
    }

    public void clear () {
        _store.clear ();
    }

    public void clear (Object minPriority) {
        //todo: optimize lookup to log(n) --> AccessStrategy.firstGreaterEquals
        final Iterator iter = _accessStrategy.begin (_store);
        while (iter.hasCurrent () && !passesThreshold (((Pair) iter.current ()).getFirst (), minPriority))
            iter.next ();
        while (iter.hasCurrent ())
            iter.remove ();
    }

    public void setAcceptanceThreshold (Object o) {
        if (o != null) {
            clear (o);
        }
        _threshold = o;
    }

    public boolean isThreadSafe () {
        return _isThreadSafe;
    }
}
