package ahc.collection;

import ahc.collection.store.Store;
import ahc.collection.access.AccessStrategy;
import ahc.util.SynchronizedGetCode;
import ahc.exception.TimeoutException;

import java.util.NoSuchElementException;


/**
 * This is the generic, configurable implementation of a queue.
 *
 * @author Arno Haase
 */
public class GenericQueue implements Queue {
    private final Store _store;
    private final AccessStrategy _accessStrategy;
    private final boolean _isThreadSafe;


    final SynchronizedGetCode _removeCode;
    final SynchronizedGetCode _peekCode;


    public GenericQueue (Store store, AccessStrategy accessStrategy, boolean isThreadSafe) {
        if (!store.preservesOrdering ())
            throw new IllegalArgumentException ("the underlying store of a queue must preserve ordering");

        _store = store;
        _accessStrategy = accessStrategy;
        _isThreadSafe = isThreadSafe;


        _removeCode = new SynchronizedGetCode (store) {
            public boolean precondition () {
                return _store.size () > 0;
            }

            public Object getUnsynchronized () {
                if (isEmpty ())
                    throw new NoSuchElementException ("queue is empty");

                final Iterator iter = _accessStrategy.end (_store);
                final Object result = iter.current ();
                iter.remove ();
                return result;
            }
        };


        _peekCode = new SynchronizedGetCode (_store) {
            public boolean precondition () {
                return _store.size () > 0;
            }

            public Object getUnsynchronized () {
                if (isEmpty ())
                    throw new NoSuchElementException ("queue is empty");

                return _accessStrategy.end (_store).current ();
            }
        };
    }


    public void add (Object obj) {
        if (_isThreadSafe) {
            synchronized (_store) {
                _accessStrategy.prepend (_store, obj);
                _store.notify ();
            }
        }
        else
            _accessStrategy.prepend (_store, obj);
    }

    public Object remove () {
        if (_isThreadSafe) {
            throw new IllegalStateException ();
        }
        else
            return _removeCode.getUnsynchronized ();
    }


    public Object remove (boolean blocking) throws InterruptedException {
        if (_isThreadSafe) {
            return _removeCode.get (blocking);
        }
        else
            throw new IllegalStateException ("makes sense only for thread safe queues");
    }


    public Object remove (int timeoutMillis) throws InterruptedException, TimeoutException {
        if (_isThreadSafe) {
            return _removeCode.get (timeoutMillis);
        }
        else
            throw new IllegalStateException ("makes sense only for thread safe queues");
    }


    public Object peek () {
        if (_isThreadSafe) {
            throw new IllegalStateException ();
        }
        else
            return _peekCode.getUnsynchronized ();
    }


    public Object peek (boolean blocking) throws InterruptedException {
        if (_isThreadSafe) {
            return _peekCode.get (blocking);
        }
        else
            throw new IllegalStateException ("makes sense only for thread safe queues");
    }

    public Object peek (int timeoutMillis) throws InterruptedException, TimeoutException {
        if (_isThreadSafe) {
            return _peekCode.get (timeoutMillis);
        }
        else
            throw new IllegalStateException ("makes sense only for thread safe queues");
    }

    public int size () {
        if (_isThreadSafe) {
            synchronized (_store) {
                return _store.size ();
            }
        }
        else
            return _store.size ();
    }

    public boolean isEmpty () {
        if (_isThreadSafe) {
            synchronized (_store) {
                return _store.isEmpty ();
            }
        }
        else
            return _store.isEmpty ();
    }

    public void clear () {
        if (_isThreadSafe) {
            synchronized (_store) {
                _store.clear ();
            }
        }
        else
            _store.clear ();
    }

    public boolean isThreadSafe () {
        return _isThreadSafe;
    }
}
