package ahc.collection;

import ahc.collection.store.Store;
import ahc.collection.store.ArrayStore;
import ahc.collection.equality.EqualityStrategy;
import ahc.collection.equality.EqualsEqualityStrategy;
import ahc.collection.access.AccessStrategy;
import ahc.collection.access.TransparentAccessStrategy;

import java.util.EmptyStackException;


/**
 * This is the generic, configurable implementation of a Stack.
 *
 * @author Arno Haase
 */
public class GenericStack implements Stack {
    private static final EqualityStrategy _eqStrat = new EqualsEqualityStrategy ();
    private static final AccessStrategy _transparentAccess = TransparentAccessStrategy.INSTANCE;

    private final Store _store;
    private final AccessStrategy _accessStrategy;


    public GenericStack (Store store, AccessStrategy accessStrategy) {
        if (!store.preservesOrdering ())
            throw new IllegalArgumentException ("The store for a stack must preserve element ordering");
        _store = store;
        _accessStrategy = accessStrategy;
    }

    public GenericStack (int initialSize) {
        _store = new ArrayStore (_eqStrat, initialSize);
        _accessStrategy = _transparentAccess;
    }

    public GenericStack () {
        _store = new ArrayStore (_eqStrat);
        _accessStrategy = _transparentAccess;
    }


    public void push (Object o) {
        _accessStrategy.add (_store, o);
    }


    public Object pop () {
        if (size() < 1)
            throw new EmptyStackException();

        final Iterator iter = _accessStrategy.end (_store);
        final Object result = iter.current ();
        iter.remove ();
        return result;
    }

    public Object peek () {
        return _accessStrategy.end (_store).current ();
    }

    public boolean isEmpty () {
        return _store.isEmpty ();
    }

    public int size () {
        return _store.size ();
    }

    public void clear () {
        _store.clear ();
    }

    public Object pop (int n) {
        if (size() < n)
            throw new EmptyStackException();

        Object result = null;
        for (int i = 0; i < n; i++)
            result = pop ();
        return result;
    }

    public Object peek (int n) {
        final Iterator iter = _accessStrategy.end (_store);
        iter.previous (n);
        return iter.current ();
    }
}
