package ahc.collection;

import ahc.collection.access.AccessStrategy;
import ahc.collection.access.GenericAccessStrategy;
import ahc.collection.equality.EqualityStrategy;
import ahc.collection.equality.EqualsEqualityStrategy;
import ahc.collection.equality.IdentityEqualityStrategy;
import ahc.collection.store.Store;
import ahc.collection.store.HashStore;
import ahc.collection.store.DlinkedStore;
import ahc.collection.store.ArrayStore;
import ahc.collection.GenericMap;
import ahc.collection.Map;

import java.util.Comparator;


/**
 * This class provides conveninence methods for the creation and management of Map instances.
 *
 * @author Arno Haase
 */
public class MapHelper {
    public static final EqualityStrategy EQ_EQUALS = new EqualsEqualityStrategy ();
    public static final EqualityStrategy EQ_IDENTITY = new IdentityEqualityStrategy ();

    public static final AccessStrategy ACCESS_UNIQUE = GenericAccessStrategy.create (null, true);
    public static final AccessStrategy ACCESS_MULTIPLE = GenericAccessStrategy.create (null, false);


    //----------------------- Helper Methods -----------------------------

    public static Map synchronizedMap (Map map) {
        if (map == null)
            return null;

        return SynchronizedMap.create (map);
    }

    //----------------------- createMap -----------------------------

    public static Map createMap () {
        return createHashMap();
    }

    //----------------------- createArrayMap ------------------------


    public static Map createArrayMap () {
        return createArrayMap (EQ_EQUALS, ACCESS_UNIQUE);
    }

    public static Map createArrayMap (int initialSize) {
        return createArrayMap (EQ_EQUALS, ACCESS_UNIQUE, initialSize);
    }

    public static Map createArrayMap (EqualityStrategy eqStrat) {
        return createArrayMap (eqStrat, ACCESS_UNIQUE);
    }

    public static Map createArrayMap (EqualityStrategy eqStrat, int initialSize) {
        return createArrayMap (eqStrat, ACCESS_UNIQUE, initialSize);
    }

    public static Map createArrayMap (Comparator comp) {
        return createArrayMap (EQ_EQUALS, GenericAccessStrategy.create (new EntryKeyComparator(comp), true));
    }

    public static Map createArrayMap (Comparator comp, int initialSize) {
        return createArrayMap (EQ_EQUALS, GenericAccessStrategy.create (new EntryKeyComparator(comp), true), initialSize);
    }

    public static Map createArrayMap (AccessStrategy accessStrat) {
        return createArrayMap (EQ_EQUALS, accessStrat);
    }

    public static Map createArrayMap (AccessStrategy accessStrat, int initialSize) {
        return createArrayMap (EQ_EQUALS, accessStrat, initialSize);
    }

    public static Map createArrayMap (EqualityStrategy eqStrat, AccessStrategy accessStrat) {
        return createMap (new ArrayStore (eqStrat), accessStrat);
    }

    public static Map createArrayMap (EqualityStrategy eqStrat, AccessStrategy accessStrat, int initialSize) {
        return createMap (new ArrayStore (eqStrat, initialSize), accessStrat);
    }


    //----------------------- createDlinkedMap ------------------------


    public static Map createDlinkedMap () {
        return createDlinkedMap (EQ_EQUALS, ACCESS_UNIQUE);
    }

    public static Map createDlinkedMap (EqualityStrategy eqStrat) {
        return createDlinkedMap (eqStrat, ACCESS_UNIQUE);
    }

    public static Map createDlinkedMap (Comparator comp) {
        return createDlinkedMap (EQ_EQUALS, GenericAccessStrategy.create (new EntryKeyComparator(comp), true));
    }

    public static Map createDlinkedMap (AccessStrategy accessStrat) {
        return createDlinkedMap (EQ_EQUALS, accessStrat);
    }

    public static Map createDlinkedMap (EqualityStrategy eqStrat, AccessStrategy accessStrat) {
        return createMap (new DlinkedStore (eqStrat), accessStrat);
    }


    //----------------------- createHashMap ------------------------


    public static Map createHashMap () {
        return createHashMap (EQ_EQUALS);
    }

    public static Map createHashMap (int initialSize) {
        return createHashMap (EQ_EQUALS, initialSize);
    }

    public static Map createHashMap (int initialSize, double loadFactor) {
        return createHashMap (EQ_EQUALS, initialSize, loadFactor);
    }

    public static Map createHashMap (EqualityStrategy eqStrat) {
        return createHashMap (eqStrat, ACCESS_UNIQUE);
    }

    public static Map createHashMap (EqualityStrategy eqStrat, int initialSize) {
        return createHashMap (eqStrat, ACCESS_UNIQUE, initialSize);
    }

    public static Map createHashMap (EqualityStrategy eqStrat, int initialSize, double loadFactor) {
        return createHashMap (eqStrat, ACCESS_UNIQUE, initialSize, loadFactor);
    }

    public static Map createHashMap (AccessStrategy accessStrat) {
        return createHashMap (EQ_EQUALS, accessStrat);
    }

    public static Map createHashMap (AccessStrategy accessStrat, int initialSize) {
        return createHashMap (EQ_EQUALS, accessStrat, initialSize);
    }

    public static Map createHashMap (AccessStrategy accessStrat, int initialSize, double loadFactor) {
        return createHashMap (EQ_EQUALS, accessStrat, initialSize, loadFactor);
    }

    public static Map createHashMap (EqualityStrategy eqStrat, AccessStrategy accessStrat) {
        return createMap (new HashStore (eqStrat), accessStrat);
    }

    public static Map createHashMap (EqualityStrategy eqStrat, AccessStrategy accessStrat, int initialSize) {
        return createMap (new HashStore (eqStrat, initialSize), accessStrat);
    }

    public static Map createHashMap (EqualityStrategy eqStrat, AccessStrategy accessStrat, int initialSize, double loadFactor) {
        return createMap (new HashStore (eqStrat, initialSize, loadFactor), accessStrat);
    }


    //----------------------- generic createMap ------------------------


    public static Map createMap (Store store, AccessStrategy accessStrat) {
        return new GenericMap (store, accessStrat);
    }

    //todo: create convenience shortcuts that provide a separate equalityStrategy for values?
    public static Map createMap (Store store, AccessStrategy accessStrat, EqualityStrategy valueEqualityStrat) {
        return new GenericMap (store, accessStrat, valueEqualityStrat);
    }
}
