package ahc.collection;

import ahc.exception.TimeoutException;


/**
 * A PriorityQueue is a queue in which all elements have an object signifying
 *  its priority; it is up to implementations to specify a valid range of
 *  objects and a means of ordering them.
 *
 * <p>Elements are returned based on their priority, highest priority is returned
 *  first. Elements of same priority are returned in the same order they were
 *  added.
 *
 * @author Arno Haase
 */
public interface PriorityQueue {
    boolean add (Object obj, Object priority);

    /**
     * This method is implemented only for non-threadsafe queues so that it does
     *  not have InterruptedException in its signature. Therefore, this method does
     *  not block but rather throws NoSuchElementException if called on an empty
     *  queue.
     */
    Object remove ();
    Object remove (boolean blocking) throws InterruptedException;
    Object remove (int timeoutMillis) throws InterruptedException, TimeoutException;

    /**
     * same as remove(), but ignores elements of a lower priority than a minimum threshold.
     */
    Object remove (Object minPriority);
    Object remove (Object minPriority, boolean blocking) throws InterruptedException;
    Object remove (Object minPriority, int timeoutMillis) throws InterruptedException, TimeoutException;

    /**
     * This method is implemented only for non-threadsafe queues so that it does
     *  not have InterruptedException in its signature. Therefore, this method does
     *  not block but rather throws NoSuchElementException if called on an empty
     *  queue.
     */
    Object peek ();
    Object peek (boolean blocking) throws InterruptedException;
    Object peek (int timeoutMillis) throws InterruptedException, TimeoutException;

    /**
     * same as remove(), but ignores elements of a lower priority than a minimum threshold.
     */
    Object peek (Object minPriority);
    Object peek (Object minPriority, boolean blocking) throws InterruptedException;
    Object peek (Object minPriority, int timeoutMillis) throws InterruptedException, TimeoutException;


    int size ();
    int size (Object minPriority);
    boolean isEmpty ();
    boolean isEmpty (Object minPriority);

    void clear ();
    /**
     * removes all elements from the queue that have a lower priority than the specified
     *  threshold.
     */
    void clear (Object minPriority);

    /**
     * Tells the queue to discard elements that have a lower priority than o. Passing in "null"
     *  clears the acceptance threshold, i.e. all elements are accepted afterwards.
     */
    void setAcceptanceThreshold (Object o);

    boolean isThreadSafe ();
}
