package ahc.collection.access;

import ahc.collection.Iterator;
import ahc.collection.access.AccessStrategy;
import ahc.collection.store.Store;

import java.util.Comparator;


/**
 * @author Arno Haase
 */
public class GenericAccessStrategy implements AccessStrategy {
    private final boolean _filterDuplicates;
    private final Comparator _comparator;

    //todo: subclasses for different store implementations

    public static AccessStrategy create (Comparator comp, boolean filterDuplicates) {
        return new GenericAccessStrategy (comp, filterDuplicates);
    }

    protected GenericAccessStrategy (boolean filterDuplicates) {
        this (null, filterDuplicates);
    }

    public GenericAccessStrategy (Comparator comp, boolean filterDuplicates) {


        _comparator = comp;
        _filterDuplicates = filterDuplicates;
    }


    public boolean add (Store store, Object o) {
        //todo: reuse this lookup for sorted insert
        if (_filterDuplicates && first (store, o) != null)
            return false;

        if (_comparator == null) {
            store.append (o);
            return true;
        }

        //todo: optimize algorithm for the different store types
        Iterator iter = store.begin ();
        while (iter.hasCurrent() && _comparator.compare (o, iter.current ()) > 0)
            iter.next ();
        iter.add (o);
        return true;
    }

    public boolean prepend (Store store, Object o) {
        //todo: reuse this lookup for sorted insert
        if (_filterDuplicates && first (store, o) != null)
            return false;

        if (_comparator == null) {
            store.prepend (o);
            return true;
        }

        //todo: optimize algorithm for the different store types
        Iterator iter = store.begin ();
        while (iter.hasCurrent() && _comparator.compare (o, iter.current ()) > 0)
            iter.next ();
        iter.add (o);
        return true;
    }

    public boolean append (Store store, Object o) {
        //todo: reuse this lookup for sorted insert
        if (_filterDuplicates && first (store, o) != null)
            return false;

        if (_comparator == null) {
            store.append (o);
            return true;
        }

        //todo: optimize algorithm for the different store types
        Iterator iter = store.end ();
        Iterator previous = store.after ();
        while (iter.hasCurrent() && _comparator.compare (o, iter.current ()) < 0) {
            previous.previous ();
            iter.previous ();
        }

        previous.add (o);
        return true;
    }


    public boolean remove (Store store, Object o) {
        //todo: optimize lookup for sorted stores
        return store.remove (o);
    }

    public Object get (Store store, int index) {
        return store.get (index);
    }

    public boolean add (Store store, Iterator iter, Object o) {
        if (_comparator != null)
            throw new UnsupportedOperationException ("modification by index is not supported for sorted collections");

        if (_filterDuplicates && first (store, o) != null)
            return false;

        iter.add (o);
        return true;
    }

    public Iterator before (Store store) {
        return new SafeIterator (store.before (), store);
    }

    public Iterator begin (Store store) {
        return new SafeIterator (store.begin (), store);
    }

    public Iterator end (Store store) {
        return new SafeIterator (store.end (), store);
    }

    public Iterator after (Store store) {
        return new SafeIterator (store.after (), store);
    }

    public Iterator iter (Store store, int index) {
        return new SafeIterator (store.iter (index), store);
    }

    public Iterator first (Store store, Object o) {
        final Iterator inner = store.first (o);
        return (inner != null) ? new SafeIterator (inner, store) : null;
    }

    public Iterator last (Store store, Object o) {
        final Iterator inner = store.last (o);
        return (inner != null) ? new SafeIterator (inner, store) : null;
    }

    private class SafeIterator implements Iterator {
        private final Iterator _inner;
        private final Store _store;

        public SafeIterator (Iterator inner, Store store) {
            _inner = inner;
            _store = store;
        }

        public boolean replace (Object o) {
            if (_filterDuplicates && first (_store, o) != null) {
                return false;
            }

            // todo: optimize
            remove ();
            add (o);
            previous();

            return true;
        }

        public void set (Object o) {
            replace (o);
        }


        public void add (Object o) {
            //todo: optimize
            if (_filterDuplicates && first (_store, o) != null) {
                return;
            }

            if (_comparator == null || _store.isSorted ()) {
                _inner.add (o);
                return;
            }

            GenericAccessStrategy.this.add (_store, o);
        }


        public boolean hasCurrent () {
            return _inner.hasCurrent();
        }

        public Object current () {
            return _inner.current ();
        }

        public int currentIndex () {
            return _inner.currentIndex ();
        }

        public Object next (int n) {
            return _inner.next (n);
        }

        public Object previous (int n) {
            return _inner.previous (n);
        }

        public boolean hasNext () {
            return _inner.hasNext ();
        }

        public Object next () {
            return _inner.next ();
        }

        public boolean hasPrevious () {
            return _inner.hasPrevious ();
        }

        public Object previous () {
            return _inner.previous ();
        }

        public int nextIndex () {
            return _inner.nextIndex ();
        }

        public int previousIndex () {
            return _inner.previousIndex ();
        }

        public void remove () {
            _inner.remove ();
        }
    }
}
