package ahc.collection.store;

import ahc.collection.equality.EqualityStrategy;
import ahc.collection.Iterator;
import ahc.collection.store.AbstractStore;

import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;


/**
 * @author Arno Haase
 */
public class ArrayStore extends AbstractStore {
    private Object[] _data;

    private int _size = 0;
    private int _modCount = 0;


    public ArrayStore (EqualityStrategy eqStrat) {
        this (eqStrat, 10);
    }

    public ArrayStore (EqualityStrategy eqStrat, int initialSize) {
        super (eqStrat);
        _data = new Object[initialSize];
    }


    public boolean preservesOrdering () {
        return true;
    }

    public boolean isSorted () {
        return false;
    }

    private void _add (int ind, Object o) {
        ensureCapacity (_size + 1);
        System.arraycopy (_data, ind, _data, ind + 1, _size - ind);
        _data[ind] = o;
        _size++;
        _modCount++;

    }


    public void prepend (Object o) {
        _add (0, o);
    }

    public void append (Object o) {
        ensureCapacity (_size + 1);
        _data[_size] = o;
        _size++;
        _modCount++;
    }

    public int size () {
        return _size;
    }

    public void clear () {
        for (int i = 0; i < _size; i++) {
            _data[i] = null; // to avoid a memory leak
        }
        _size = 0;
    }

    public Object get (int index) {
        return _data[index];
    }

    public Iterator before () {
        return iter (-1);
    }

    public Iterator begin () {
        return iter (0);
    }

    public Iterator end () {
        return iter (size () - 1);
    }

    public Iterator after () {
        return iter (size ());
    }


    public Iterator iter (int ind) {
        return new Itr (ind);
    }


    private void ensureCapacity (int minCapacity) {
        int oldCapacity = _data.length;
        if (minCapacity > oldCapacity) {
            _modCount++;
            final Object oldData[] = _data;
            int newCapacity = (oldCapacity * 3) / 2 + 1;
            if (newCapacity < minCapacity)
                newCapacity = minCapacity;
            _data = new Object[newCapacity];
            System.arraycopy (oldData, 0, _data, 0, _size);
        }
    }


    public void compactSize () {
        if (_size == _data.length)
            return;

        final Object[] newData = new Object[_size];
        System.arraycopy (_data, 0, newData, 0, _size);
        _data = newData;
        _modCount++;
    }


    private class Itr implements Iterator {
        int _curPos = -1;
        int _expectedModCount = _modCount;

        public Itr (int curPos) {
            _curPos = curPos;
        }

        public void add (Object o) {
            if (_curPos < 0 || _curPos > _size)
                throw new NoSuchElementException ("iterator pos " + _curPos + ", size is " + _size);
            checkCoModification ();

            ArrayStore.this._add (_curPos, o);
            _curPos++;

            _expectedModCount = _modCount;
        }

        public boolean hasCurrent () {
            return _curPos >= 0 && _curPos < _size;
        }

        public Object current () {
            checkRange ();
            checkCoModification ();

            return _data[_curPos];
        }

        public int currentIndex () {
            return _curPos;
        }

        public boolean replace (Object o) {
            checkCoModification ();

            _data[_curPos] = o;
            _modCount++;
            _expectedModCount = _modCount;

            return true;
        }

        public void set (Object o) {
            replace (o);
        }

        public boolean hasNext () {
            return _curPos < _size - 1;
        }

        public Object next () {
            _curPos++;

            return hasCurrent () ? current () : null;
        }

        public Object next (int n) {
            _curPos += n;
            return hasCurrent () ? current () : null;
        }

        public int nextIndex () {
            return _curPos + 1;
        }


        public boolean hasPrevious () {
            return _curPos > 0;
        }

        public Object previous () {
            _curPos--;
            return hasCurrent () ? current () : null;
        }

        public Object previous (int n) {
            _curPos -= n;
            return hasCurrent () ? current () : null;
        }

        public int previousIndex () {
            return _curPos - 1;
        }

        public void remove () {
            if (_curPos == -1)
                throw new IllegalStateException ();

            checkCoModification ();

            System.arraycopy (_data, _curPos + 1, _data, _curPos, _size - _curPos - 1);
            _size--;
            _data[_size] = null; // to remove the reference
            _modCount++;
            _expectedModCount = _modCount;
        }

        private void checkRange () {
            if (!hasCurrent ())
                throw new NoSuchElementException ("iterator pos " + _curPos + ", size is " + _size);
        }

        private void checkCoModification () {
            if (_expectedModCount != _modCount)
                throw new ConcurrentModificationException ();
        }
    }
}
