package ahc.collection.store;

import ahc.collection.Iterator;
import ahc.collection.store.AbstractStore;
import ahc.collection.equality.EqualityStrategy;

import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;


/**
 * @author Arno Haase
 */
public class DlinkedStore extends AbstractStore {
    private final Entry _before = new Entry (null, null, null);
    private final Entry _after = new Entry (null, null, null);

    private int _size = 0;
    private int _modCount = 0;


    public DlinkedStore (EqualityStrategy eqStrat) {
        super (eqStrat);

        _before._next = _after;
        _after._prev = _before;
    }


    public boolean preservesOrdering () {
        return true;
    }

    public boolean isSorted () {
        return false;
    }

    public void prepend (Object o) {
        final Entry newElement = new Entry (o, _before, _before._next);
        _before._next._prev = newElement;
        _before._next = newElement;

        _size++;
        _modCount++;
    }

    public void append (Object o) {
        final Entry newElement = new Entry (o, _after._prev, _after);
        _after._prev._next = newElement;
        _after._prev = newElement;

        _size++;
        _modCount++;
    }

    public Object get (int index) {
        return entry (index)._element;
    }

    public int size () {
        return _size;
    }

    public void clear () {
        _before._next = _after;
        _after._prev = _before;
        _size = 0;
    }

    private Entry entry (int ind) {
        if (ind <= _size / 2) {
            Entry entry = _before;

            for (int i = 0; i <= ind; i++) {
                entry = entry._next;
            }

            return entry;
        }
        else {
            Entry entry = _after;

            for (int i = 0; i < _size - ind; i++) {
                entry = entry._prev;
            }

            return entry;
        }
    }

    public Iterator iter (int ind) {
        return new Itr (entry (ind), ind);
    }


    public Iterator before () {
        return new Itr (_before, -1);
    }

    public Iterator begin () {
        return new Itr (_before._next, 0);
    }

    public Iterator end () {
        return new Itr (_after._prev, size () - 1);
    }

    public Iterator after () {
        return new Itr (_after, size ());
    }


    public void compactSize () {
        // do nothing: a dlinked list can not be compacted
    }


    private static class Entry {
        Object _element;
        Entry _prev;
        Entry _next;

        public Entry (Object element, Entry prev, Entry next) {
            _element = element;
            _prev = prev;
            _next = next;
        }
    }


    private class Itr implements Iterator {
        Entry _curPos = _before;
        int _expectedModCount = _modCount;
        int _curIndex;

        public Itr (Entry curPos, int curIndex) {
            _curPos = curPos;
            _curIndex = curIndex;
        }

        public void add (Object o) {
            if (_curPos == _before)
                throw new NoSuchElementException ();
            checkCoModification ();

            final Entry newElement = new Entry (o, _curPos._prev, _curPos);
            _curPos._prev._next = newElement;
            _curPos._prev = newElement;
            _curIndex++;
            _size++;

            _modCount++;
            _expectedModCount = _modCount;
        }

        public boolean hasCurrent () {
            return _curPos != _after && _curPos != _before;
        }

        public Object current () {
            checkRange ();
            checkCoModification ();

            return _curPos._element;
        }

        public int currentIndex () {
            return _curIndex;
        }


        public boolean replace (Object o) {
            checkRange ();
            checkCoModification ();

            _curPos._element = o;

            _modCount++;
            _expectedModCount = _modCount;
            return true;
        }

        public void set (Object o) {
            replace (o);
        }

        public boolean hasNext () {
            return _curPos._next != _after && _curPos != _after;
        }

        public Object next () {
            _curPos = _curPos._next;
            _curIndex++;

            return hasCurrent () ? current () : null;
        }

        public Object next (int n) {
            for (int i = 0; i < n; i++)
                next ();
            return hasCurrent () ? current () : null;
        }

        public int nextIndex () {
            return _curIndex + 1;
        }


        public boolean hasPrevious () {
            return _curPos._prev != _before && _curPos != _before;
        }

        public Object previous () {
            _curPos = _curPos._prev;
            _curIndex--;
            return hasCurrent () ? current () : null;
        }

        public Object previous (int n) {
            for (int i = 0; i < n; i++)
                previous ();
            return hasCurrent () ? current () : null;
        }

        public int previousIndex () {
            return _curIndex - 1;
        }


        public void remove () {
            checkRange ();
            checkCoModification ();

            _curPos._prev._next = _curPos._next;
            _curPos._next._prev = _curPos._prev;
            _curPos = _curPos._next;
            _size--;
        }

        private void checkRange () {
            if (!hasCurrent ())
                throw new NoSuchElementException ();
        }

        private void checkCoModification () {
            if (_expectedModCount != _modCount)
                throw new ConcurrentModificationException ();
        }
    }
}
