package ahc.collection.store;

import ahc.collection.equality.EqualityStrategy;
import ahc.collection.store.AbstractStore;
import ahc.collection.Iterator;

import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;


/**
 * @author Arno Haase
 */
public class HashStore extends AbstractStore {

    private Entry[] _buckets;
    private final double _loadFactor;
    private int _threshold;

    private int _size = 0;
    private int _modCount = 0;

    //todo: implement optimized version of first()



    public HashStore (EqualityStrategy eqStrat) {
        this (eqStrat, 10);
    }

    public HashStore (EqualityStrategy eqStrat, int initialSize) {
        this (eqStrat, initialSize, 0.75);
    }

    public HashStore (EqualityStrategy eqStrat, int initialSize, double loadFactor) {
        super (eqStrat);
        _loadFactor = loadFactor;
        _threshold = (int) (initialSize * loadFactor);
        _buckets = new Entry[initialSize];
    }

    public boolean preservesOrdering () {
        return false;
    }

    public boolean isSorted () {
        return false;
    }


    public void prepend (Object o) {
        add (o);
    }

    private void add (Object o) {
        ensureCapacity ();
        final int indBucket = getBucket (o);
        _buckets[indBucket] = new Entry (o, _buckets[indBucket]);
        _size++;
        _modCount++;
    }

    public void append (Object o) {
        add (o);
    }

    public int size () {
        return _size;
    }

    public void clear () {
        _buckets = new Entry[_buckets.length];
        _size = 0;
    }

    public Iterator before () {
        return new Itr ();
    }

    public Iterator begin () {
        final Iterator result = new Itr ();
        result.next ();
        return result;
    }

    public Iterator end () {
        throw new UnsupportedOperationException ();
    }

    public Iterator after () {
        throw new UnsupportedOperationException ();
    }

    public Iterator iter (int ind) {
        throw new UnsupportedOperationException ();
    }

    private int getBucket (Object o, int numBuckets) {
        return (_eqStrat.hashCode (o) & 0x7fffffff) % numBuckets;
    }

    private int getBucket (Object o) {
        return getBucket (o, _buckets.length);
    }


    public Iterator first (Object o) {
        int indBucket = getBucket (o);
        Entry e = _buckets[indBucket];
        while (e != null) {
            if (_eqStrat.equals (e._element, o)) {
                return new Itr (indBucket, e);
            }
            e = e._next;
        }

        return null;
    }

    //todo: optimized version for "last"

    public void compactSize () {
        rehash ((int) (size () / _loadFactor) + 1);
    }

    private void ensureCapacity () {
        if (_size > _threshold)
            rehash (_buckets.length * 2 + 1);
    }


    public void rehash () {
        rehash (_buckets.length);
    }

    public void rehash (int numBuckets) {
        _modCount++;
        final Entry[] newBuckets = new Entry[numBuckets];

        for (Iterator iter = before (); iter.hasNext (); ) {
            iter.next ();
            final int indBucket = getBucket (iter.current (), numBuckets);
            newBuckets[indBucket] = new Entry (iter.current (), newBuckets[indBucket]);
        }

         _threshold = (int) (numBuckets * _loadFactor);
        _buckets = newBuckets;
    }

    private static class Entry {
        Object _element;
        Entry _next;

        public Entry (Object element, Entry next) {
            _element = element;
            _next = next;
        }
    }

    private class Itr implements Iterator {
        private int _expectedModCount = _modCount;

        private int _indBucket = 0;
        private Entry _curEntry = null;
        private boolean _isAfter = false;


        public Itr () {
        }

        private Itr (int indBucket, Entry curEntry) {
            _indBucket = indBucket;
            _curEntry = curEntry;
        }

        public void add (Object o) {
            throw new UnsupportedOperationException ();
        }

        public boolean hasCurrent () {
            return _curEntry != null && !_isAfter;
        }

        public Object current () {
            checkRange ();
            checkCoModification ();

            return _curEntry._element;
        }

        public int currentIndex () {
            throw new UnsupportedOperationException ();
        }

        public boolean replace (Object o) {
            throw new UnsupportedOperationException ();
        }

        public void set (Object o) {
            replace (o);
        }


        public boolean hasNext () {
            int baseInd = _indBucket;

            if (_curEntry != null) {
                if (_curEntry._next != null)
                    return true;
                else
                    baseInd++;
            }

            for (int ind = baseInd; ind < _buckets.length; ind++) {
                if (_buckets[ind] != null)
                    return true;
            }

            return false;
        }


        public Object next () {
            checkCoModification ();

            if (_curEntry != null) {
                if (_curEntry._next != null) {
                    _curEntry = _curEntry._next;
                    return _curEntry._element;
                }
                else {
                    _indBucket++;
                    _curEntry = null;
                }
            }

            for (int ind = _indBucket; ind < _buckets.length; ind++) {
                if (_buckets[ind] != null) {
                    _indBucket = ind;
                    _curEntry = _buckets[ind];
                    return _curEntry._element;
                }
            }

            _isAfter = true;
            return null;
        }

        public Object next (int n) {
            for (int i = 0; i < n; i++)
                next ();
            return current ();
        }

        public int nextIndex () {
            throw new UnsupportedOperationException ();
        }


        public boolean hasPrevious () {
            throw new UnsupportedOperationException ();
        }

        public Object previous () {
            throw new UnsupportedOperationException ();
        }

        public Object previous (int n) {
            throw new UnsupportedOperationException ();
        }

        public int previousIndex () {
            throw new UnsupportedOperationException ();
        }


        public void remove () {
            checkRange ();
            checkCoModification ();

            Entry prev = _buckets[_indBucket];
            if (prev == _curEntry) {
                _buckets[_indBucket] = _curEntry._next;
            }
            else {
                while (prev._next != _curEntry) {
                    prev = prev._next;
                }
                prev._next = _curEntry._next;
            }

            _size--;
            _modCount++;
            // this operation invalidates the iterator

            _indBucket = 0;
            _curEntry = null;
        }


        private void checkRange () {
            if (!hasCurrent ())
                throw new NoSuchElementException ();
        }

        private void checkCoModification () {
            if (_expectedModCount != _modCount)
                throw new ConcurrentModificationException ();
        }
    }
}
