package ahc.generic;


/**
 * This interface serves to retrieve attributes from an object by successively calling get methods. The
 *  methods that are to be called are encoded in the "accessPath" parameter.<p>
 *
 * For example,
 * <pre>
 * attributeGetter.get (person, "address/city/name")
 * </pre> is equivalent to
 * <pre>
 * person.getAddress().getCity().getName()
 * </pre>
 *
 * This sort of programming should be avoided when possible because it prevents compile time checks for
 *  consistency and is slower than directly calling the attributes. It is however sometimes helpful when writing
 *  generic code that should handle different situations and not know about the details of concrete business
 *  classes, e.g. in a framework.<p>
 *
 * Note that such an approach can only work if <em>all</em> get methods that are called on the way are
 *  <code>public</code>, otherwise the generic code doing the lookup will not be able to "see" them and an
 *  exception will result.<p>
 *
 * Several implementations of this interface are available; see
 * <a href="AttributeGetterFactory.html">AttributeGetterFactory</a> for details.
 *
 * @author Arno Haase
 */
public interface AttributeGetter {
    String CHAR_SEPARATOR = "/";

    public Object get (Object obj, String accessPath);
}
