package ahc.generic;

import ahc.generic.classloader.ClassLoaderAttributeGetter;


/**
 * This class serves as a factory for different kinds of AttributeGetter implementations. The purpose of an
 *  AttributeGetter is to retrieve arbitrary attributes from the objects that are reachable via get methods
 *  from the original object; for a more detailed description see
 *  <a href="AttributeGetter.html">AttributeGetter</a>.<p>
 *
 * Traditionally, such lookups are done using reflection. That has the disadvantage of being very slow. So this
 *  factory offers three different implementations with different trade-offs - see the descriptions of the
 *  actual create methods for details.<p>
 *
 * The following list contains the results of one performance measurement I made. These results should of course
 *  be taken with a grain of salt, but they give a rough idea of what to expect for long running applications.
 *  The results are durations, so less is better.<p>
 *
 * <table border="0">
 * <tr><td>classloader implementations:</td><td align="right">18</td></tr>
 * <tr><td>caching reflection: </td><td align="right">190</td></tr>
 * <tr><td>non-caching reflection:</td><td align="right">2120</td></tr>
 * </table>
 *
 * @author Arno Haase
 */
public class AttributeGetterFactory {
    /**
     * This method creates and returns an AttributeGetter implementation that is based on a custom classloader
     *  implementation. This implemntation is by far the fastest, and if in doubt this is probably the
     *  one that you want. <p>
     * A specific class is created (and cached) for every type of object and every access path. That leads to
     *  several specific aspects of this implementation:<p>
     * <ul>
     * <li> Creating the access classes incurs some performance penalty. The long term benefit by far outweighs
     *       this penalty, but you should keep it in mind.<br> Note that the performance penalty is primarily in terms
     *       of the time it takes the JIT compiler to optimize the newly generated classes. The delay for
     *       actually generating the classes is negligible, but it takes a couple of calls to the methods
     *       until the code is fully optimized.<p>
     * <li> Caching: Every call to this create method creates a new instance of the AttributeGetter
     *       implementation, with a new and separate instance of the cache. This is done to give the client
     *       application control over the growth of the cache. But it means that the client application
     *       <em>must</em> keep hold of a reference to the AttributeGetter implementation, otherwise there
     *       will be no caching.<p>
     * <li> Cache size: Often, this sort of "lookup" of values on objects using a string representation of the
     *       requested attribute is done for a quite limited number of classes and attributes. In this case the
     *       size of the cache should be negligible for most applications (it is in the order of 1 kB per class
     *       and attribute). Still, it should be checked if these assumptions hold for the concrete client -
     *       the cache of the implementation is never released until the AttributeGetter instance is garbage
     *       collected.
     * <li> The EJB specification officially prohibits - together with many other features of Java - the use
     *       of custom classloaders. If you are a stickler for formal rules, you may want to consider the "caching
     *       reflection" version in an EJB setting.
     * </ul>
     */
    public static AttributeGetter createClassLoaderGetter () {
        return new ClassLoaderAttributeGetter ();
    }

    /**
     * This method creates a straightforward implementation that uses reflection and does no caching whatsoever.<p>
     *  As indicated above, this approach is several orders of magnitude slower than the classloader approach, so
     *  the only reason to use it is to save memory by avoiding caching.
     */
    public static AttributeGetter createNonCachingReflectionGetter () {
        return new NonCachingGetter ();
    }


    /**
     * This implementation uses reflection but performs aggressive caching of intermediate results. The issues
     *  listed for caching of the classloader approach are valid here as well.<p>
     * This implementation is usually far slower than the classloader implementation and uses significantly more
     *  cache memory, but optimization by the JIT compiler happens faster. The only context I am aware of where
     *  you might want to use this implementation is in an EJB container where the spec prohibits the use of
     *  custom class loaders.
     */
    public static AttributeGetter createCachingReflectionGetter () {
        return new CachingGetter ();
    }
}
