package ahc.io;

import java.io.*;


/**
 * The purpose of this class is to act as a cache between a read-only file (e.g. a
 *  configuration file) and its read object representation.<br>
 *
 * Clients access the object representation of the object through the <code>getData</code>
 *  method. That checks the timestamp of the underlying file to determine if it has changed
 *  since it was last read. In the common case that it has not changed, a cached copy of
 *  the object representing the file is returned, and only when necessary the  <code>read</code>
 *  method provided by the subclass is called to actually access the file.<br>
 *
 * The <code>minReadInterval</code> parameter serves as a "security buffer" in case the
 *  file is changed incrementally so that the file's time stamp is changed before the
 *  changes are complete and the file is again consistent. The file is reread only when
 *  this interval has passed since the last change to the timestamp of the file.
 *
 * @author Arno Haase
 */
public abstract class FileWatcher {
    private long _lastRead = 0;
    private long _lastModified = 0;
    private Object _cached = null;

    private final File _file;
    private final long _minRereadInterval;


    public FileWatcher (String filename) {
        this (filename, 1000);
    }


    public FileWatcher (String filename, long minRereadInterval) {
        _file = new File (filename);
        _minRereadInterval = minRereadInterval;
    }


    public synchronized Object getData () throws IOException {
        final long now = System.currentTimeMillis();

        if (_cached == null ||
                (now - _lastRead > _minRereadInterval && _lastModified != _file.lastModified ())) {
            _lastRead = now;
            _lastModified = _file.lastModified();
            _cached = read (new BufferedInputStream (new FileInputStream (_file)));
        }

        return _cached;
    }


    protected abstract Object read (InputStream is) throws IOException;
}
