package ahc.io;

import ahc.util.bytes.InputStreamToByteArray;

import javax.swing.*;
import java.io.InputStream;
import java.io.IOException;


/**
 * This class serves to retrieve resources from the class path, i.e. it allows clients to access the contents
 *  of files (e.g. images) that are in the classpath. This is often better than reading from the file system
 *  because it reduces installation hassle, works in application servers etc.
 *
 * @author Arno Haase
 */
public class ResourceLoader {
    /**
     * This is the underlying method of this class; it provides access to a resource. The resource name
     *  must be provided relative to the class path root and with '/' as separators, e.g.
     *  <code>"images/MyImage.jpg"</code> for an image in the subdirectory "images", or
     *  <code>"ahc/util/ResourceLoader.class</code> for the class file of this class.<p>
     *
     * If the resource is not found <code>null</code> is returned.
     */
    public static InputStream getAsStream (String name) {
        return Thread.currentThread ().getContextClassLoader ().getResourceAsStream (name);
    }


    /**
     * This is a convenience method that reads an entire resource and returns it as a byte array.<p>
     *
     * If the resource is not found, an IllegalArgumentException is thrown.
     */
    public static byte[] getAsBytes (String name) {
        final InputStream is = getAsStream (name);

        if (is == null)
            throw new IllegalArgumentException ("resource not found: '" + name + "'");

        try {
            return new InputStreamToByteArray (1024).read (getAsStream (name));
        } catch (IOException exc) {
            throw new RuntimeException ("should not have happened: " + exc);
        }
    }


    /**
     * This is a convenience method that reads an entire resource and returns it as an ImageIcon.<p>
     *
     * If the resource is not found, an IllegalArgumentException is thrown.
     */
    public static ImageIcon getAsImageIcon (String name) {
        return new ImageIcon (getAsBytes (name));
    }
}
