package ahc.jdbc.logger;

import java.lang.reflect.Proxy;
import java.sql.*;
import java.util.Properties;


/**
 * This class is a JDBC driver that acts as a wrapper around an arbitrary other JDBC driver and logs all
 *  calls made to that driver to <code>System.out</code>. It can be used in any way that other JDBC drivers
 *  can be used.<p>
 *
 * The URL is <code>jdbclog:&lt;driver class&gt;:&lt;driver URL&gt;</code>
 * <ul>
 * <li> driver class: the fully qualified class name of the "real" JDBC driver to which the calls are forwarded
 *       and that actually executes them.
 * <li> driver URL: the URL that you would normally use for the JDBC driver. For DB/2, this could be something
 *       like jdbc:db2:test
 * </ul>
 *
 * Putting it all together, the URL could be something like
 * <code>jdbclog:COM.ibm.db2.jdbc.app.DB2Driver:jdbc:db2:test</code>.<p>
 *
 * In the logging output, every "real" object that is returned by a JDBC call (i.e. something like a Connection
 *  or a ResultSet rather than a String, Date or Timestamp) is associated with a unique ID. Whenever a call is
 *  made, the ID is logged for the object on which the call is made. That allows you to trace what calls
 *  are made on which objects in which sequence.<p>
 *
 * Note: This logging is <em>very</em> verbose. It is meant for tracking down some rather nasty bugs. When you
 *  use it, make sure that you redirect System.out to a file.
 *
 * @author Arno Haase
 */
public class LoggingJdbcDriver implements Driver {
    public static final String URL_PREFIX = "jdbclog:";

    static {
        try {
            DriverManager.registerDriver (new LoggingJdbcDriver ());
        } catch (SQLException exc) {
            //TODO: Logging
        }
    }


    public Connection connect (String url, Properties info) throws SQLException {
        final String urlWithoutPrefix = url.substring (URL_PREFIX.length ());

        final int indColon = urlWithoutPrefix.indexOf (":");
        if (indColon < 0)
            return null;

        final String driverClassName = urlWithoutPrefix.substring (0, indColon);
        final String innerUrl = urlWithoutPrefix.substring (indColon + 1);

        try {
        final Driver driver = (Driver) Class.forName(driverClassName).newInstance();

        return (Connection) Proxy.newProxyInstance (getClass ().getClassLoader (),
                new Class[]{Connection.class},
                new LoggingJdbcInvocationHandler (Connection.class,
                        driver.connect (innerUrl, info),
                        UniqueIdFactory.createUniqueId ()));
        }
        catch (ClassNotFoundException exc) {
            exc.printStackTrace();
            return null;
        }
        catch (InstantiationException exc) {
            exc.printStackTrace();
            return null;
        }
        catch (IllegalAccessException exc) {
            exc.printStackTrace();
            return null;
        }
    }

    public boolean acceptsURL (String url) throws SQLException {
        return url.startsWith (URL_PREFIX);
    }

    public DriverPropertyInfo[] getPropertyInfo (String url, Properties info) throws SQLException {
        return new DriverPropertyInfo[0];
    }

    public int getMajorVersion () {
        return 1;
    }

    public int getMinorVersion () {
        return 0;
    }

    public boolean jdbcCompliant () {
        return true;
    }
}
