package ahc.jdbc.logger;

import ahc.util.string.StringHelper;
import ahc.util.string.ToString;

import java.lang.reflect.*;
import java.sql.*;
import java.util.HashMap;
import java.util.Map;


class LoggingJdbcInvocationHandler implements InvocationHandler {
    private final Class _type;
    private final Object _inner;
    private final String _uniqueId;

    public LoggingJdbcInvocationHandler (Class type, Object inner, String uniqueId) {
        _type = type;
        _inner = inner;
        _uniqueId = uniqueId;
    }

    public Object invoke (Object proxy, Method method, Object[] args) throws Throwable {
        try {
            final InvocationResult result = actualInvokeMethod (method, args);

            logCall (method, args, result._result, result._uniqueId);

            return result.getResult ();
        } catch (Throwable th) {
            if (th instanceof InvocationTargetException)
                th = ((InvocationTargetException) th).getTargetException ();

            logCall (method, args, th);
            throw th;
        }
    }


    private static class InvocationResult {
        private final Object _result;
        private final boolean _hasUniqueId;
        private final String _uniqueId;

        public InvocationResult (Object result, boolean hasUniqueId, String uniqueId) {
            this._result = result;
            _hasUniqueId = hasUniqueId;
            _uniqueId = uniqueId;
        }

        public Object getResult () {
            return _result;
        }

        public boolean isHasUniqueId () {
            return _hasUniqueId;
        }

        public String getUniqueId () {
            return _uniqueId;
        }
    }


    private void logCall (Method method, Object[] args, Object result, String uniqueId) {
        if (filterLogCalls (method))
            return;

        System.out.println ("{jdbclog}" +
                "[" + _uniqueId + "][" + System.currentTimeMillis () + "] " +
                _type.getName () + "." + method.getName () +
                " (" + ToString.defaultToString(args) + ") " +
                ": " +
                ((result instanceof String) ? "\"" + result + "\"" : StringHelper.escape ("" + result)) +
                ((uniqueId != null) ? "[" + uniqueId + "]" : ""));
        //TODO: Logging einbauen
    }

    private boolean filterLogCalls (Method method) {
        if ("toString".equals (method.getName ()) &&
                method.getParameterTypes ().length == 0)
            return true;

        return false;
    }

    private void logCall (Method method, Object[] args, Throwable th) {
        System.out.println ("{jdbclog}" +
                "[" + _uniqueId + "][" + System.currentTimeMillis () + "] " +
                _type.getName () + "." + method.getName () +
                " (" + ToString.defaultToString(args) + ") " +
                ": Exception " + StringHelper.escape ("" + th));
        //TODO: Logging einbauen
    }


    private static final Map _specialReturnTypes = new HashMap ();

    static {
        _specialReturnTypes.put (CallableStatement.class, LoggingJdbcInvocationHandler.class);
        _specialReturnTypes.put (Connection.class, LoggingJdbcInvocationHandler.class);
        _specialReturnTypes.put (DatabaseMetaData.class, LoggingJdbcInvocationHandler.class);
        _specialReturnTypes.put (PreparedStatement.class, LoggingJdbcInvocationHandler.class);
        _specialReturnTypes.put (ResultSet.class, LoggingJdbcInvocationHandler.class);
        _specialReturnTypes.put (ResultSetMetaData.class, LoggingJdbcInvocationHandler.class);
        _specialReturnTypes.put (Statement.class, LoggingJdbcInvocationHandler.class);
    }


    private InvocationResult actualInvokeMethod (Method method, Object[] args)
            throws IllegalAccessException, InvocationTargetException, NoSuchMethodException, InstantiationException {
        final Object result = method.invoke (_inner, args);
        final Class returnType = method.getReturnType ();

        if (_specialReturnTypes.containsKey (returnType)) {
            final String uniqueId = UniqueIdFactory.createUniqueId ();

            final Object resultObject = Proxy.newProxyInstance (getClass ().getClassLoader (),
                    new Class[]{returnType},
                    getHandlerInstance ((Class) _specialReturnTypes.get (returnType), returnType, result, uniqueId));

            return new InvocationResult (resultObject, true, uniqueId);
        }

        return new InvocationResult (result, false, null);
    }

    private InvocationHandler getHandlerInstance (Class handlerClass, Class wrappedInterface, Object inner, String uniqueId)
            throws NoSuchMethodException, IllegalAccessException, InstantiationException, InvocationTargetException {
        final Constructor ctor = handlerClass.getConstructor (new Class[]{Class.class, Object.class, String.class});
        return (InvocationHandler) ctor.newInstance (new Object[]{wrappedInterface, inner, uniqueId});
    }
}
