package ahc.util;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.io.Serializable;


/**
 * This class represents a date, i.e. a combination of day, month and year.
 *
 * <p>It fills a gap in the JDK libraries: java.util.Date is a point in time including hours, minutes, secondes etc.,
 *  and it does a reasonably good job of it. A point in time is prone to time zone conversions etc. so that
 *  its representation depends on a time zone / locale setting.
 *
 * <p>Dates are different. If some one's birthday is Sept 5 1974, then that is it, and we usually do <em>not</em>
 *  want the birthday to be displayed differently for different time zones. In a business setting it is also quite
 *  commonly necessary to do date arithmetic: one week later, one month previously etc. The common approach of
 *  treating a day as 24 * 60 * 60 * 1000 milliseconds and adding / subtracting these is not only cumbersome but
 *  also a violation of the Date contract (there is such a thing as leap seconds...) Doing this sort of thing
 *  cleanly requires the use of GregorianCalendar which is complex, cumbersome and expensive to use.
 *
 * <p>To cut a long story short, this class fills the gap by providing a small, inexpensive class for date
 *  representation and arithmetic. The class is immutable, and all modifying methods return new instances and
 *  leave the original instance unmodified.
 *
 * @author Arno Haase
 */
public class DateDay implements Comparable, Serializable {
    private final int _day, _month, _year;

    private static final int[] MIN_NUM_DAYS_IN_MONTH = {-1, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
    private static final int[] MAX_NUM_DAYS_IN_MONTH = {-1, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

    private static final int NUM_MONTHS = 12;

    private DateDay (int day, int month, int year) {
        if (month < 1 || month > NUM_MONTHS)
            throw new IllegalArgumentException ("illegal month " + month + ", must be in range from 1 to 12");

        if (day < 1 || day > numDays (month, year))
            throw new IllegalArgumentException ("illegal day " + day + " for month " + month + " in year " + year + ".");

        _day = day;
        _month = month;
        _year = year;
    }

    /** returns a DateDay instance based on day, month and year */
    public static DateDay create (int day, int month, int year) {
        return new DateDay (day, month, year);
    }

    /** returns a DateDay instance based on the full java.util representation, i.e. a Date and a TimeZone */
    public static DateDay create (Date date, TimeZone timezone) {
        if (date == null)
            return null;

        final Calendar cal = new GregorianCalendar (timezone);
        cal.setTime (date);

        return new DateDay (cal.get (Calendar.DAY_OF_MONTH), cal.get (Calendar.MONTH) + 1, cal.get (Calendar.YEAR));
    }


    /** returns a DateDay instance based on a java.util.Date and the default TimeZone */
    public static DateDay create (Date date) {
        return create (date, TimeZone.getDefault ());
    }


    public int getDay () {
        return _day;
    }

    public int getMonth () {
        return _month;
    }

    public int getYear () {
        return _year;
    }

    public DateDay setDay (int day) {
        return create (day, _month, _year);
    }

    public DateDay setMonth (int month) {
        return create (_day, month, _year);
    }

    public DateDay setYear (int year) {
        return create (_day, _month, year);
    }


    /** converts to a java.util.Date based on the default TimeZone */
    public Date asDate () {
        return asDate (TimeZone.getDefault ());
    }

    /** converts to a java.util.Date based on a specific TimeZone */
    public Date asDate (TimeZone timeZone) {
        final Calendar cal = new GregorianCalendar (timeZone);
        cal.clear ();
        cal.set (Calendar.DAY_OF_MONTH, _day);
        cal.set (Calendar.MONTH, _month - 1);
        cal.set (Calendar.YEAR, _year);

        return cal.getTime ();
    }


    /** implements java.util.Comparable. This method expects the other object to be an instance of DateDay.*/
    public int compareTo (Object o) {
        final DateDay other = (DateDay) o;

        if (other._year != _year)
            return _year - other._year;

        if (other._month != _month)
            return _month - other._month;

        return _day - other._day;
    }

    public boolean before (DateDay other) {
        return compareTo (other) < 0;
    }

    public boolean after (DateDay other) {
        return compareTo (other) > 0;
    }


    public static boolean isLeapYear (int year) {
        if (year % 4 != 0)
            return false;

        if (year % 400 == 0)
            return true;

        return year % 100 != 0;
    }

    public static int numDays (int month, int year) {
        return isLeapYear (year) ? MAX_NUM_DAYS_IN_MONTH[month] : MIN_NUM_DAYS_IN_MONTH[month];
    }

    /**
     * returns the next date
     */
    public DateDay incDay () {
        return normalizedDay (_day + 1, _month, _year);
    }

    /**
     * returns the equivalent of <em>num<em> calls to indDay()
     */
    public DateDay incDay (int num) {
        return normalizedDay (_day + num, _month, _year);
    }

    private DateDay normalizedDay (int day, int month, int year) {
        if (month <= 0)
            return normalizedDay (day, month + NUM_MONTHS, year - 1);
        if (month > NUM_MONTHS)
            return normalizedDay (day, month - NUM_MONTHS, year + 1);

        if (day > numDays (month, year))
            return normalizedDay (day - numDays (month, year), month + 1, year);
        if (day <= 0) {
            month--;
            if (month == 0) {
                month = NUM_MONTHS;
                year--;
            }
            return normalizedDay (day + numDays (month, year), month, year);
        }

        return create (day, month, year);
    }

    private DateDay normalizedMonth (int day, int month, int year, boolean roundLater) {
        while (month < 1) {
            month += NUM_MONTHS;
            year--;
        }
        while (month > NUM_MONTHS) {
            month -= NUM_MONTHS;
            year++;
        }

        if (day <= numDays (month, year))
            return create (day, month, year);

        return roundLater ?
                normalizedDay (1, month + 1, year) :
                create (numDays (month, year), month, year);
    }

    /**
     * returns the previous date
     */
    public DateDay decDay () {
        return normalizedDay (_day - 1, _month, _year);
    }

    /**
     * returns the equivalent of <em>num</em> calls to decDay
     */
    public DateDay decDay (int num) {
        return normalizedDay (_day - num, _month, _year);
    }

    /** returns the same day of the next month. If the current day of the month is bigger than the maximum
     *  for that month, the parameter flag roundLater determines whether the first of the month after (true) or
     *  the last day of that month is used (false).
     */
    public DateDay incMonth (boolean roundLater) {
        return incMonth (1, roundLater);
    }

    /** returns the same day <em>num</em> months later. If the current day of the month is bigger than the maximum
     *  for that month, the parameter flag roundLater determines whether the first of the month after (true) or
     *  the last day of that month is used (false).
     *
     * <p> This is <em>not</em> equivalent to <em>num</em> calls to incMonth(boolean)!
     */
    public DateDay incMonth (int num, boolean roundLater) {
        return normalizedMonth (_day, _month + num, _year, roundLater);
    }


    /**
     * The opposite of incMonth(boolean)
     */
    public DateDay decMonth (boolean roundLater) {
        return decMonth (1, roundLater);
    }

    /**
     * The opposite of incMonth (int, boolean)
     */
    public DateDay decMonth (int num, boolean roundLater) {
        return normalizedMonth (_day, _month - num, _year, roundLater);
    }


    /**
     * returns the same date of the next year. In the rare case that that year does not have that date
     *  (i.e. 29th Feb), the parameter flag roundLater determines wheter the previous date (false) or the
     *  next date (true) is returned.
     */
    public DateDay incYear (boolean roundLater) {
        return incYear (1, roundLater);
    }

    /**
     * returns the same date <em>num</em> years later. In the rare case that that year does not have that date
     *  (i.e. 29th Feb), the parameter flag roundLater determines wheter the previous date (false) or the
     *  next date (true) is returned.
     *
     * <p> This is <em>not</em> equivalent to several subsequent calls of incYear (boolean)!
     */
    public DateDay incYear (int num, boolean roundLater) {
        return normalizedMonth(_day, _month, _year + num, roundLater);
    }


    /**
     * The opposite of incYear(boolean)
     */
    public DateDay decYear (boolean roundLater) {
        return decYear (1, roundLater);
    }

    /**
     * The opposite of incYear(int, boolean)
     */
    public DateDay decYear (int num, boolean roundLater) {
        return normalizedMonth(_day, _month, _year - num, roundLater);
    }


    public String toString () {
        return asDate ().toString ();
    }

    public boolean equals (Object o) {
        if (this == o) return true;
        if (!(o instanceof DateDay)) return false;

        final DateDay dateDay = (DateDay) o;

        if (_day != dateDay._day) return false;
        if (_month != dateDay._month) return false;
        if (_year != dateDay._year) return false;

        return true;
    }

    public int hashCode () {
        int result;
        result = _day;
        result = 29 * result + _month;
        result = 29 * result + _year;
        return result;
    }
}


//todo: diff, nextWorkday