package ahc.util;


/**
 * This class is an implementation of the GOF "Flyweight" pattern. It provides a single point of
 *  access to Integer objects and caches them. Since Integer objects are immutable, an entire
 *  program can e.g. share a single instance of e.g. "Integer (1)"; this class caches instances
 *  and thus allows memory to be saved.<br>
 * This class uses a fixed number of 1000 bins to store the integers (analogously to a HashMap)
 *  in order to maximize retrieval performance. If it is used for numbers larger than that, it
 *  will still work but performance will degrade due to the linear lookup overhead inside each
 *  bin.<br>
 * CAUTION: This class never releases an Integer once it was created. So this class should be
 *  used only in those cases when that is desired, e.g. for counters of relatively small numbers
 *  in collections.
 *
 * @author Arno Haase
 */
public class IntegerFactory {
    public static final int NUM_BINS = 1000;

    private static final Entry[] _bins = new Entry[NUM_BINS];


    public static Integer get (int value) {
        final int indBin = getBin (value);
        Entry entry = _bins[indBin];

        if (entry == null) {
            final Integer result = new Integer (value);
            _bins[indBin] = new Entry (result, null);
            return result;
        }

        // this is just to make the compiler happy. The loop is executed at least
        // once, so this value is overwritten.
        Entry previous = null;

        while (entry != null) {
            if (entry._element.intValue () == value)
                return entry._element;

            previous = entry;
            entry = entry._next;
        }

        final Integer result = new Integer (value);
        final Entry newEntry = new Entry (result, null);
        previous._next = newEntry;

        return result;
    }

    private static int getBin (int value) {
        return (value & 0x7fffffff) % NUM_BINS;
    }

    /**
     * convenience method for retrieving the next higher number. A parameter value of null is treated
     *  as 0.
     */
    public static Integer increment (Integer value) {
        return get (getValueNotNull (value) + 1);
    }

    private static int getValueNotNull (Integer value) {
        return (value == null) ? 0 : value.intValue ();
    }

    /**
     * convenience method for retrieving the next lower number. A parameter value of null is treated
     *  as 0.
     */
    public static Integer decrement (Integer value) {
        return get (getValueNotNull (value) - 1);
    }

    /**
     * convenience method for retrieving the sum. A parameter value of null is treated
     *  as 0.
     */
    public static Integer add (Integer v1, Integer v2) {
        return get (getValueNotNull (v1) + getValueNotNull (v2));
    }

    /**
     * convenience method for retrieving the sum. A parameter value of null is treated
     *  as 0.
     */
    public static Integer add (Integer v1, int v2) {
        return get (getValueNotNull (v1) + v2);
    }


    /**
     * convenience method for retrieving the difference. A parameter value of null is treated
     *  as 0.
     */
    public static Integer subtract (Integer v1, Integer v2) {
        return get (getValueNotNull (v1) - getValueNotNull (v2));
    }

    /**
     * convenience method for retrieving the difference. A parameter value of null is treated
     *  as 0.
     */
    public static Integer subtract (Integer v1, int v2) {
        return get (getValueNotNull (v1) - v2);
    }

    /**
     * convenience method for retrieving the difference. A parameter value of null is treated
     *  as 0.
     */
    public static Integer subtract (int v1, Integer v2) {
        return get (v1 - getValueNotNull (v2));
    }


    /** an implementation of the mathematically typical - as opposed to the Java - implementation
     *  of the modulus funtion. For "a mod b", this implementation returns a number c such that <br>
     *
     * <ul>
     *   <li> 0 <= c < b, and
     *   <li> a - c is a whole multiple of b,
     * </ul>
     *
     * This differs from the usual Java implementation of a % b which returns negative numbers for
     *  negative a.
     */
    public static Integer mod (int a, int b) {
        return get (IntHelper.mod (a, b));
    }

    /** an implementation of the mathematically typical - as opposed to the Java - implementation
     *  of the modulus funtion. For "a mod b", this implementation returns a number c such that <br>
     *
     * <ul>
     *   <li> 0 <= c < b, and
     *   <li> a - c is a whole multiple of b,
     * </ul>
     *
     * This differs from the usual Java implementation of a % b which returns negative numbers for
     *  negative a.
     */
    public static Integer mod (Integer a, int b) {
        return mod (a.intValue (), b);
    }

    /** an implementation of the mathematically typical - as opposed to the Java - implementation
     *  of the modulus funtion. For "a mod b", this implementation returns a number c such that <br>
     *
     * <ul>
     *   <li> 0 <= c < b, and
     *   <li> a - c is a whole multiple of b,
     * </ul>
     *
     * This differs from the usual Java implementation of a % b which returns negative numbers for
     *  negative a.
     */
    public static Integer mod (int a, Integer b) {
        return mod (a, b.intValue ());
    }

    /** an implementation of the mathematically typical - as opposed to the Java - implementation
     *  of the modulus funtion. For "a mod b", this implementation returns a number c such that <br>
     *
     * <ul>
     *   <li> 0 <= c < b, and
     *   <li> a - c is a whole multiple of b,
     * </ul>
     *
     * This differs from the usual Java implementation of a % b which returns negative numbers for
     *  negative a.
     */
    public static Integer mod (Integer a, Integer b) {
        return mod (a.intValue (), b.intValue ());
    }


    private static class Entry {
        public Entry (Integer element, Entry next) {
            _element = element;
            _next = next;
        }

        Integer _element;
        Entry _next;
    }
}
