package ahc.util;

import ahc.exception.TimeoutException;


/**
 * This class and performs a blocking wait on a lock and then retrieves an object.<br>
 *
 * The get operation first waits until a precondition is met and it can acquire a monitor, or until a
 *  timout is reached. The monitor is passed in to the constructor, and the precondition must be provided
 *  by implementing an abstract method.<br>
 *
 * If the precondition is met and the lock acquired, the actuel result is retrieved by calling the abstract
 *  hook "getUnsynchronized".
 *
 * This implementation retrieves the result without passing any parameters to the getUnsynchronized method.
 *
 * @author Arno Haase
 */
public abstract class SynchronizedGetCode {

    private final SynchronizedParamGetCode _inner;

    public SynchronizedGetCode (Object lock) {
        _inner = new SynchronizedParamGetCode (lock) {
            public boolean precondition (Object parameter) {
                return SynchronizedGetCode.this.precondition ();
            }

            public Object getUnsynchronized (Object parameter) {
                return SynchronizedGetCode.this.getUnsynchronized ();
            }
        };
    }


    public Object get (int timeoutMillis) throws InterruptedException, TimeoutException {
        return _inner.get (null, timeoutMillis);
    }

    public Object get (boolean blocking) throws InterruptedException {
        return _inner.get (null, blocking);
    }


    public abstract boolean precondition ();
    public abstract Object getUnsynchronized ();
}
