package ahc.util;

import ahc.exception.TimeoutException;


/**
 * This class and performs a blocking wait on a lock and then retrieves an object.<br>
 *
 * The get operation first waits until a precondition is met and it can acquire a monitor, or until a
 *  timout is reached. The monitor is passed in to the constructor, and the precondition must be provided
 *  by implementing an abstract method.<br>
 *
 * If the precondition is met and the lock acquired, the actuel result is retrieved by calling the abstract
 *  hook "getUnsynchronized".
 *
 * This implementation retrieves the result passing one parameter to the getUnsynchronized method.
 *
 * @author Arno Haase
 */
public abstract class SynchronizedParamGetCode {
    private final Object _lock;

    public SynchronizedParamGetCode (Object lock) {
        _lock = lock;
    }

    private void waitForPrecondition (Object parameter, long timeoutMillis)
            throws InterruptedException, TimeoutException {
        final long timeoutLimit = System.currentTimeMillis () + timeoutMillis;
        final long waitInterval = timeoutMillis / 5;

        while (!precondition (parameter) && System.currentTimeMillis () < timeoutLimit)
            _lock.wait (waitInterval);

        if (!precondition (parameter))
            throw new TimeoutException (timeoutMillis);
    }

    private void waitForPrecondition (Object parameter) throws InterruptedException {
        while (!precondition (parameter))
            wait ();
    }


    public Object get (Object parameter, int timeoutMillis) throws InterruptedException, TimeoutException {
        if (timeoutMillis < 0)
            return get (parameter, true);

        synchronized (_lock) {
            waitForPrecondition (parameter, timeoutMillis);
            return getUnsynchronized (parameter);
        }
    }

    public Object get (Object parameter, boolean blocking) throws InterruptedException {
        synchronized (_lock) {
            waitForPrecondition (parameter);
            return getUnsynchronized (parameter);
        }
    }

    public abstract boolean precondition (Object parameter);

    public abstract Object getUnsynchronized (Object parameter);
}
