package ahc.util;

import java.util.Comparator;


/**
 * This class provides a decorator for another Comparator that allows inverting the direction of the
 *  resulting sorting.<p>
 *
 * For example, a normal Collator provides lexical ordering (placing 'a' before 'b' etc.). Using this
 *  decorator class, it becomes possible to switch between this normal ordering and its inverse (placing 'b'
 *  before 'a').<p>
 *
 * This class is particularly useful in the context of user interfaces where the user may want to control
 *  the sort order of a table interactively.
 *
 * @author Arno Haase
 */
public class VarDirectionComparator implements Comparator {
    private final Comparator _inner;
    private boolean _isDirInverted = false;

    public VarDirectionComparator (Comparator inner) {
        _inner = inner;
    }

    public VarDirectionComparator (Comparator inner, boolean dirInverted) {
        _inner = inner;
        _isDirInverted = dirInverted;
    }

    public int compare (Object o1, Object o2) {
        return _isDirInverted ? -_inner.compare (o1, o2) : _inner.compare (o1, o2);
    }


    public boolean isDirInverted () {
        return _isDirInverted;
    }

    public void setDirInverted (boolean dirInverted) {
        _isDirInverted = dirInverted;
    }

    public void toggleDirection () {
        _isDirInverted = !_isDirInverted;
    }
}
