package ahc.util.string;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.*;


/**
 * @author Arno Haase
 */
abstract class AttribIterator {
    protected final boolean _filterStatic;
    protected final boolean _filterTransient;
    protected final boolean _alwaysUseToStringImpl;
    protected final boolean _collectionsAsPrimitives;
    protected final boolean _javaLangAsPrimitives;

    private final Map _cycles = new HashMap (); // identityHashCode (o) -> cycle no.
    private int _nextCycleNo = 1;

    private static final Object _firstOccurrence = new Object () {
        public String toString () {
            return "<once>";
        }
    };


    public AttribIterator (boolean filterStatic, boolean filterTransient, boolean alwaysUseToStringImpl,
            boolean collectionsAsPrimitives, boolean javaLangAsPrimitives) {
        _filterStatic = filterStatic;
        _filterTransient = filterTransient;
        _alwaysUseToStringImpl = alwaysUseToStringImpl;
        _collectionsAsPrimitives = collectionsAsPrimitives;
        _javaLangAsPrimitives = javaLangAsPrimitives;
    }


    protected abstract void handlePre (Object obj);

    protected abstract void handlePreCollection (Object obj);

    protected abstract void handlePost (Object obj);

    protected abstract void handlePostCollection (Object obj);

    protected abstract void handleNode (Object obj);

    protected abstract void handleEnumSeparator ();

    protected abstract void handleMapArrow ();

    protected abstract void handleField (Field field);

    protected abstract void handleCycleDef (Object obj);

    protected abstract void handleCycleRef (Object obj);

    protected abstract void handleObject (Object obj);


    public void handleGeneric (Object o, int maxDepth) {
        handleGeneric (o, false, maxDepth);
    }


    protected Object createKey (Object obj) {
        return new Integer (System.identityHashCode (obj));
    }

    void handleGeneric (Object o, boolean isPrimitive, int maxDepth) {
        handleObject (o); // for cache initialization

        if (o == null || isPrimitive || maxDepth <= 0) {
            handleAsPrimitive (o);
            return;
        }

        final Object key = createKey (o);
        final Object oldValue = _cycles.get (key);

        if (oldValue == null) {
            _cycles.put (key, _firstOccurrence);
        }
        else if (oldValue == _firstOccurrence) {
            _cycles.put (key, new Integer (_nextCycleNo++));
        }

        if (isCycle (o)) {
            handleCycleRef (o);
            return;
        }

        handleCycleDef (o);

        if (o.getClass ().isArray ()) {
            handleArray (o, maxDepth);
        }
        else if (o instanceof Collection && _collectionsAsPrimitives) {
            handleCollection ((Collection) o, maxDepth);
        }
        else if (o instanceof Map && _collectionsAsPrimitives) {
            handleMap ((Map) o, maxDepth);
        }
        else if (_javaLangAsPrimitives && String.class.getPackage().equals (o.getClass ().getPackage ())) {
            handleAsPrimitive (o);
        }
        else if (_alwaysUseToStringImpl && hasOwnToString (o.getClass ())) {
            handleAsPrimitive (o);
        }
        else {
            handlePre (o);

            final Field[] fields = (Field[]) getFields (o.getClass ()).toArray (new Field[0]);
            for (int i = 0; i < fields.length; i++) {
                if (i > 0)
                    handleEnumSeparator ();

                try {
                    fields[i].setAccessible (true);
                    handleField (fields[i]);

                    final Object curValue = fields[i].get (o);
                    handleGeneric (curValue, fields[i].getType ().isPrimitive (), maxDepth - 1);
                }
                catch (IllegalAccessException e) {
                    throw new IllegalStateException ("illegal access for field " + fields[i]);
                }
            }

            handlePost (o);
        }
    }


    private void handleArray (Object o, int maxDepth) {
        handlePreCollection (o);

        final boolean isPrimitive = o.getClass ().getComponentType ().isPrimitive ();
        final int len = Array.getLength (o);
        for (int i = 0; i < len; i++) {
            if (i > 0)
                handleEnumSeparator ();

            final Object element = Array.get (o, i);
            handleGeneric (element, isPrimitive, maxDepth);
        }

        handlePostCollection (o);
    }

    private void handleCollection (Collection coll, int maxDepth) {
        handlePreCollection (coll);

        boolean first = true;
        for (Iterator iter = coll.iterator (); iter.hasNext ();) {
            if (!first)
                handleEnumSeparator ();
            first = false;

            handleGeneric (iter.next (), false, maxDepth);
        }

        handlePostCollection (coll);
    }

    private void handleMap (Map map, int maxDepth) {
        handlePreCollection (map);

        boolean first = true;
        for (Iterator iter = map.keySet ().iterator (); iter.hasNext ();) {
            if (!first)
                handleEnumSeparator ();
            first = false;

            final Object key = iter.next ();
            handleGeneric (key, false, maxDepth);
            handleMapArrow ();
            handleGeneric (map.get (key), false, maxDepth);
        }

        handlePostCollection (map);
    }

    private void handleAsPrimitive (Object o) {
        if (o == null) {
            handleNode (null);
        }
        else if (o.getClass ().isArray ()) {
            handleArray (o, 0);
        }
        else if (o instanceof Collection) {
            handleCollection ((Collection) o, 0);
        }
        else if (o instanceof Map) {
            handleMap ((Map) o, 0);
        }
        else {
            handleNode (o);
        }
    }


    private boolean hasOwnToString (Class cls) {
        try {
            Method mtdToString = cls.getMethod ("toString", null);
            return mtdToString.getDeclaringClass ().equals (cls);
        }
        catch (Exception e) {
        }
        return false;
    }


    private List getFields (Class cls) {
        final List result = new ArrayList ();

        if (Object.class.equals (cls))
            return result;

        final Field[] fields = cls.getDeclaredFields ();
        for (int i = 0; i < fields.length; i++) {
            final int curModifiers = fields[i].getModifiers ();

            if (_filterStatic && Modifier.isStatic (curModifiers))
                continue;
            if (_filterTransient && Modifier.isTransient (curModifiers))
                continue;

            result.add (fields[i]);
        }

        result.addAll (getFields (cls.getSuperclass ()));
        return result;
    }

    protected boolean isCycle (Object obj) {
        return getCycleNo (obj) != null && getCycleNo (obj) != _firstOccurrence;
    }

    protected Object getCycleNo (Object obj) {
        return _cycles.get (createKey (obj));
    }
}
