package ahc.util.string;

import ahc.util.bytes.ByteArray;

import java.io.UnsupportedEncodingException;


/**
 * This class provides methods for encoding arbitrary data (i.e. byte arrays) in Base64, a standardized
 *  format using only "printable" characters.
 *
 * @author Arno Haase
 */
public class Base64Formatter {
    private static final char FILL_CHAR = '=';
    private static final String ENC_ISO_8859_1 = "iso-8859-1";

    /**
     * encodes the given data and returns the result in string form. <p>
     *
     * Passing in <code>null</code> return <code>null</code>.<p>
     *
     * Strictly speaking, the result of the encoding process is a byte array (or rather, a sequence of ASCII
     *  characters, which is basically the same) and not a string so that the string representation would
     *  depend on the encoding used. But given
     *  the nature of Base64, the string representation of the characters used is widely independent of the
     *  encoding used as long as it is a single-byte encoding. Specifically, ASCII, ISO-8859-1 and UTF-8 all
     *  yield the same results.
     */
    public static String encodeToString (byte[] data) {
        final byte[] result = encode (data);
        if (result == null)
            return null;

        try {
            return new String (result, ENC_ISO_8859_1);
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException ("'" + ENC_ISO_8859_1 + "' not supported...");
        }
    }


    /**
     * encodes the given data and returns the result in string form. <p>
     *
     * Passing in <code>null</code> return <code>null</code>.<p>
     */
    public static byte[] encode (byte[] data) {
        if (data == null)
            return null;

        final ByteArray dest = new ByteArray ();

        final int destLength = ((data.length + 2) / 3) * 4;

        for (int indSource = 0; indSource < data.length; indSource += 3) {
            dest.addInt8 (expand1 (data, indSource));
            dest.addInt8 (expand2 (data, indSource));
            dest.addInt8 (expand3 (data, indSource));
            dest.addInt8 (expand4 (data, indSource));
        }

        final byte[] result = dest.asBytes ();

        // overwrite trailing bytes that were wrongly initialized above for convenience
        for (int idx = destLength - 1; idx > (data.length * 4) / 3; idx--)
            result[idx] = FILL_CHAR;

        return result;
    }


    private static int expand4 (byte[] data, int indSource) {
        return encodeAscii (getNoLimit (data, indSource + 2) & 0x3f);
    }

    private static int expand3 (byte[] data, int indSource) {
        return encodeAscii (getNoLimit (data, indSource + 2) >>> 6 & 0x03 | getNoLimit (data, indSource + 1) << 2 & 0x3f);
    }

    private static int expand2 (byte[] data, int indSource) {
        return encodeAscii (getNoLimit (data, indSource + 1) >>> 4 & 0x0f | getNoLimit (data, indSource) << 4 & 0x3f);
    }

    private static int expand1 (byte[] data, int indSource) {
        return encodeAscii (getNoLimit (data, indSource) >>> 2 & 0x3f);
    }

    private static byte getNoLimit (byte[] data, int ind) {
        if (ind < data.length)
            return data[ind];

        return 0;
    }

    private static int encodeAscii (int value) {
        if (value < 26)
            return value + 'A';
        if (value < 52)
            return value - 26 + 'a';
        if (value < 62)
            return value - 52 + '0';

        if (value == 62)
            return '+';

        return '/';
    }


    /**
     * This method decodes a sequence of bytes from Base64 encoding into the original byte sequence.<p>
     *
     * Passing in <code>null</code> returns <code>null</code>.<p>
     *
     * Strictly speaking, the result of the encoding process is a byte array (or rather, a sequence of ASCII
     *  characters, which is basically the same) and not a string so that the string representation would
     *  depend on the encoding used. But given
     *  the nature of Base64, the string representation of the characters used is widely independent of the
     *  encoding used as long as it is a single-byte encoding. Specifically, ASCII, ISO-8859-1 and UTF-8 all
     *  yield the same results.
     */
    public static final byte[] decode (String str) {
        if (str == null) {
            return null;
        }

        try {
            return decode (str.getBytes (ENC_ISO_8859_1));
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException ("'" + ENC_ISO_8859_1 + "' not supported...");
        }
    }


    private static byte decodeAscii (int value) {
        if (value >= 'A' && value <= 'Z')
            return (byte) (value - 'A');
        if (value >= 'a' && value <= 'z')
            return (byte) (value - 'a' + 26);
        if (value >= '0' && value <= '9')
            return (byte) (value - '0' + 52);
        if (value == '+')
            return 62;
        if (value == '/')
            return 63;
        if (value == FILL_CHAR)
            return 0;

        throw new IllegalArgumentException ("invalid character " + new Character ((char) value));
    }


    /**
     * This method decodes a sequence of bytes from Base64 encoding into the original byte sequence.<p>
     *
     * Passing in <code>null</code> returns <code>null</code>.
     */
    public static final byte[] decode (byte data[]) {
        if (data == null)
            return null;

        final int lenResult = indLastNonFill (data) - data.length / 4;

        final ByteArray result = new ByteArray ();
        for (int indDest = 0; indDest < lenResult; indDest += 3) {
            result.addInt8 (condense1 (data, (indDest * 4) / 3));

            if (lenResult > indDest + 1)
                result.addInt8 (condense2 (data, (indDest * 4) / 3));
            if (lenResult > indDest + 2)
                result.addInt8 (condense3 (data, (indDest * 4) / 3));
        }

        return result.asBytes ();
    }

    private static int indLastNonFill (byte[] data) {
        int result = data.length;
        while (data[result - 1] == FILL_CHAR)
            result--;

        return result;
    }

    private static byte condense3 (byte[] data, int sidx) {
        return (byte) (decodeAscii (data[sidx + 2]) << 6 & 0xff | decodeAscii (data[sidx + 3]) & 0x3f);
    }

    private static byte condense2 (byte[] data, int sidx) {
        return (byte) (decodeAscii (data[sidx + 1]) << 4 & 0xff | decodeAscii (data[sidx + 2]) >>> 2 & 0xf);
    }

    private static byte condense1 (byte[] data, int sidx) {
        return (byte) (decodeAscii (data[sidx]) << 2 & 0xff | decodeAscii (data[sidx + 1]) >>> 4 & 0x3);
    }
}
