package ahc.util.string;


/**
 * This class provides a generic toString method for arbitrary objects. It uses reflection and recursive
 *  descent to create a string representation for an object.
 *
 * <p>If there are several references to the <em>same</em> object (i.e. ref1 == ref2, not just equals) in the
 *  graph, the object is printed only once, and only a marker is inserted for further references. This also
 *  guarantees that cyclic references are handled gracefully rather than causing endless recursion.
 *
 * <p>USAGE:
 *
 * <p>The simplest way to use this class is to just call <code>ToString.defaultToString (o);</code>. That returns
 *  a default string representation of the object.
 *
 * <p>This class is however configurable to display varying levels of detail. Three preconfigured instances
 *  MIN_DETAIL, DEFAULT and MAX_DETAIL are provided as public static final variables that create minimum, common
 *  and maximum detail respectively. To create maximum detail, you would write
 *  <code>ToString.MAX_DETAIL.toString (o)</code>. Note that the instance method is <code>toString</code>
 *  whereas the static method is <code>genericToString</code>.
 *
 * <p>Even finer control over the level of detail is possible by creating a custom instance of ToString using
 *  the desired combination of flags.
 *
 * @author Arno Haase
 */
public class ToString {
    public static final ToString DEFAULT = new ToString (true, true, false, true, true, true);
    public static final ToString MIN_DETAIL = new ToString (true, true, true, true, true, true);
    public static final ToString MAX_DETAIL = new ToString (false, false, false, false, false, true);


    private final boolean _filterStatic;
    private final boolean _filterTransient;
    private final boolean _alwaysUseToStringImpl;
    private final boolean _collectionsAsPrimitives;
    private final boolean _javaLangAsPrimitives;
    private final boolean _escapeStrings;


    /**
     * returns a string representation of an object that is configured at a default level of detail.
     */
    public static String defaultToString (Object o) {
        return DEFAULT.genericToString (o, Integer.MAX_VALUE);
    }


    /**
     * Creates a ToString instance based on a custom configuration.
     *
     * @param filterStatic controls whether static variables are filtered out (true) or included (false).
     * @param filterTransient controls whether transient variables are filtered out (true) or included (false).
     * @param alwaysUseToStringImpl controls what to do if one of the objects overrides toString(). A value of
     *  true means that an overridden toString method is used, false means that the reflective algorithm is
     *  applied regardless of an overridden toString method.
     * @param collectionsAsPrimitives controls how instances of <code>java.util.Collection</code> and
     *  <code>java.util.Map</code> are handled. A value of true means that a compact default representation is
     *  used, false switches of special treatment.
     * @param javaLangAsPrimitives controls whether instances of classes from java.lang are displayed using
     *  their toString methods (true) or like any other object (false).
     * @param escapeStrings controls whether strings are displayed "as is" (false) or preprocessed so that
     *  special characters that affect layout are replaced by standard character sequences (true).
     */
    public ToString (boolean filterStatic, boolean filterTransient, boolean alwaysUseToStringImpl,
                     boolean collectionsAsPrimitives, boolean javaLangAsPrimitives, boolean escapeStrings) {
        _filterStatic = filterStatic;
        _filterTransient = filterTransient;
        _alwaysUseToStringImpl = alwaysUseToStringImpl;
        _collectionsAsPrimitives = collectionsAsPrimitives;
        _javaLangAsPrimitives = javaLangAsPrimitives;
        _escapeStrings = escapeStrings;
    }


    /**
     * returns a string representation of an object
     */
    public String toString (Object o) {
        return genericToString(o, Integer.MAX_VALUE);
    }

    /**
     * returns a string representation of an object with a limit on the depth of the recursive descent. This can
     *  e.g. be helpful to display just the vicinity of a node in a large graph.
     */
    public String genericToString (Object o, int maxDepth) {
        return createDumper ().genericToString (o, maxDepth);
    }

    /**
     * returns a string representation of an object without recursive descent. This method provides special
     *  treatment of collections and escapes strings, but otherwise the toString method is called (but in a
     *  null-safe way).
     */
    public String primitiveToString (Object o) {
        return createDumper ().genericToString(o, 0);
    }

    ToStringDumper createDumper () {
        return new ToStringDumper (_filterStatic, _filterTransient, _alwaysUseToStringImpl, _collectionsAsPrimitives, _javaLangAsPrimitives, _escapeStrings);
    }
}
