package ahc.util.timer;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.text.NumberFormat;


/**
 * This class serves to measure time in "experiments", i.e. in a test environment. The metaphor is a stop watch.<p>
 *
 * The time starts when the StopWatch is created, and it runs until "stop" is called. Calls to "current" return
 *  the current time in milliseconds since the watch was started or - if it was stopped - between the start and
 *  the stop. <p>
 *
 * It is posssible to take the time for intermediate results by calling "remember". Such a call saves the current
 *  time together with an identifying string so that it is later possible to get a list of all intermediate
 *  events with their respective times.<p>
 *
 * The <code>toString</code> method returns a human readable string representation of the entire measurement.<p>
 *
 * Note that for meaningful performance measurements it is often necessary to execute the measured code several
 *  times. <a href="TimerExperiment.html">TimerExperiment</a> provides support for that.
 *
 * @author Arno Haase
 */
public class StopWatch {
    private final String _name;

    private boolean _running = false;
    private long _startTime;
    private long _stopTime;

    private final List _events = new ArrayList ();

    private final NumberFormat _format = NumberFormat.getNumberInstance ();


    public StopWatch (String name) {
        _name = name;
        start ();
    }


    private void start () {
        _running = true;
        _startTime = System.currentTimeMillis ();
    }


    public long current () {
        if (_running)
            return System.currentTimeMillis () - _startTime;
        else
            return _stopTime - _startTime;
    }


    public void stop () {
        _running = false;
        _stopTime = System.currentTimeMillis ();
    }


    public void remember (String event) {
        if (!_running)
            throw new IllegalStateException ("Stop watch must be running to remember an event");

        _events.add (new StopWatchEvent (event, current ()));
    }


    public List getEvents () {
        return _events;
    }


    public boolean isRunning () {
        return _running;
    }

    public String getName () {
        return _name;
    }


    public String toString () {
        final StringBuffer events = new StringBuffer ();
        long timePrevEvent = 0;

        for (Iterator iter = _events.iterator (); iter.hasNext ();) {
            final StopWatchEvent curEvent = (StopWatchEvent) iter.next ();
            events.append ("  " + curEvent.getName () + ": " + _format.format (curEvent.getTime ()) + "ms (" +
                    _format.format ((curEvent.getTime () - timePrevEvent)) + "ms)\n");
            timePrevEvent = curEvent.getTime ();
        }

        return "Stopwatch \"" + _name + "\"\n" +
                events.toString () +
                "Total: " + _format.format (current ()) + "ms";
    }
}
