package ahc.util.timer;

import ahc.util.timer.StopWatch;

import java.text.NumberFormat;


/**
 * This class provides support for meaningful performance measurement of a piece of code.<p>
 *
 * To use this class, subclass it and implement the <code>doIt</code> method with the code for which the
 *  performance is to be measured. Pass the number of times that the code should be executed to the constructor.
 *  Optionally, you can also pass the number of iterations after which an intermediate result is recorded.<p>
 *
 * The name of the experiment serves to distinguish the results of several experiments that are performed in a
 *  single application.<p>
 *
 * When a piece is first executed, the VM executes it in interpreter mode which is comparatively slow. As the
 *  code is executed more and more often, the JIT compiler (or Hotspot or whatever) starts to optimize it more
 *  and more and translates it into machine code, resulting in a significant increase of speed. Using the
 *  default setting of 10 intermediate results, you can see when the execution time has stabilized (often
 *  after a couple of 1000 to 10000 calls). Performance data is usually interesting only for the time after
 *  the initial JIT compilation is finished.<p>
 *
 * The following code snippet gives an example of the usage of this class to test the performance of retrieving
 *  elements from different parts of a LinkedList:<p>
 *
 * <pre>
 * final int numIterations = 10000000;
 *
 * final List linkedList = new LinkedList ();
 * for (int i=0; i<200; i++) {
 *     linkedList.add ("Hello");
 * }
 *
 * System.out.println (new TimerExperiment ("LinkedList 0", numIterations) {
 *     public void doIt () {
 *         linkedList.get (0);
 *     }
 * }.execute());
 * System.out.println (new TimerExperiment ("LinkedList 99", numIterations) {
 *     public void doIt () {
 *         linkedList.get (99);
 *     }
 * }.execute());
 * System.out.println (new TimerExperiment ("LinkedList 199", numIterations) {
 *     public void doIt () {
 *         linkedList.get (199);
 *     }
 * }.execute());
 * </pre>
 *
 * @author Arno Haase
 */
public abstract class TimerExperiment {
    private final int _numMeasured;
    private final int _lapTimeInterval;
    private final String _name;

    private final NumberFormat _format = NumberFormat.getNumberInstance();

    public TimerExperiment (String name, int numMeasured) {
        this (name, numMeasured, numMeasured/10);
    }

    public TimerExperiment (String name, int numMeasured, int lapTimeInterval) {
        _name = name;
        _lapTimeInterval = lapTimeInterval;
        _numMeasured = numMeasured;
    }


    /**
     * performs the actual iteration and returns a StopWatch instance with the results.
     */
    public StopWatch execute () {
        final StopWatch result = new StopWatch (_name);
        for (int i=0; i<_numMeasured; i++) {
            doIt ();
            if (shouldRememberLapTime (i))
                result.remember("Time after " + _format.format(i+1) + " iterations");
        }
        result.stop();
        return result;
    }

    private boolean shouldRememberLapTime (int i) {
        return (i % _lapTimeInterval) == _lapTimeInterval - 1;
    }


    public abstract void doIt ();
}
