package ahc.collection.store;

import ahc.collection.Iterator;
import ahc.collection.store.Store;
import ahc.collection.equality.EqualityStrategy;
import ahc.collection.equality.EqualsEqualityStrategy;
import ahc.collection.equality.IdentityEqualityStrategy;
import junit.framework.TestCase;

import java.util.ArrayList;
import java.util.Arrays;


/**
 * @author Arno Haase
 */
public abstract class AbstractStoreTest extends TestCase {
    protected Store _store;
    protected final boolean _iterSupportsPrevious;

    public AbstractStoreTest (String s, boolean iterSupportsPrevious) {
        super (s);
        _iterSupportsPrevious = iterSupportsPrevious;
    }

    public void setUp () {
        _store = create (new EqualsEqualityStrategy ());

        _store.append ("a");
        _store.append ("b");
        _store.prepend ("c");
        _store.append ("d");
        _store.prepend ("a");
    }

    protected abstract Store create (EqualityStrategy eqStrat);


    protected void checkEquals (String content, Store store) {
        final Object[] ref = new Object[content.length ()];
        for (int i = 0; i < content.length (); i++) {
            ref[i] = "" + content.charAt (i);
        }

        checkEquals (ref, store.toArray ());
    }

    protected void checkEquals (Object[] arr1, Object[] arr2) {
        if (_store.preservesOrdering ())
            assertTrue (Arrays.equals (arr1, arr2));
        else {
            assertEquals (arr1.length, arr2.length);

            final java.util.List ref = new ArrayList (Arrays.asList (arr1));
            for (int i = 0; i < arr2.length; i++) {
                ref.remove (arr2[i]);
            }
            assertTrue (ref.isEmpty ());
        }
    }


    public void testEqualityStrategy () {
        final Store st1 = create (new EqualsEqualityStrategy ());
        final Store st2 = create (new IdentityEqualityStrategy ());
        st1.append (new Integer (1));
        st2.append (new Integer (1));

        assertNotNull (st1.first (new Integer (1)));
        assertNull (st2.first (new Integer (1)));
    }

    public void testRemove () {
        _store.prepend ("a");

        checkEquals ("aacabd", _store);
        assertTrue (_store.remove ("b"));
        checkEquals ("aacad", _store);
        assertTrue (_store.remove ("d"));
        checkEquals ("aaca", _store);

        assertFalse (_store.remove ("9"));

        assertTrue (_store.remove ("a"));
        checkEquals ("aca", _store);
        assertTrue (_store.remove ("c"));
        checkEquals ("aa", _store);
        assertTrue (_store.remove ("a"));
        checkEquals ("a", _store);
        assertTrue (_store.remove ("a"));
        checkEquals ("", _store);
    }


    public void testToArray () {
        checkEquals (new Object[]{"a", "c", "a", "b", "d"}, _store.toArray ());
        checkEquals (new String[]{"a", "c", "a", "b", "d"}, _store.toArray (new String[2]));
        checkEquals (new String[]{"a", "c", "a", "b", "d"}, _store.toArray (new String[5]));
        checkEquals (new String[]{"a", "c", "a", "b", "d", null, null}, _store.toArray (new String[7]));
    }

    public void testGrowth () {
        _store.compactSize ();
        for (int i = 0; i < 25; i++)
            _store.append ("x");

        assertEquals (30, _store.size ());
        checkEquals ("acabdxxxxxxxxxxxxxxxxxxxxxxxxx", _store);
    }

    private void checkIter (Store store, Iterator iter, int currentInd, Object curValue) {
        assertNotNull (iter);

        if (_store.preservesOrdering ()) {
            assertEquals (currentInd, iter.currentIndex ());
            assertEquals (currentInd - 1, iter.previousIndex ());
            assertEquals (currentInd + 1, iter.nextIndex ());

            if (currentInd >= 0 && currentInd < store.size ())
                assertEquals (curValue, iter.current ());

            assertEquals (currentInd < store.size () - 1, iter.hasNext ());
        }

        if (_iterSupportsPrevious) {
            assertEquals (currentInd > 0, iter.hasPrevious ());
        }
        assertEquals (currentInd >= 0 && currentInd < store.size (), iter.hasCurrent ());
    }

    void removeSingly (java.util.List from, java.util.List list) {
        for (java.util.Iterator iter = list.iterator (); iter.hasNext ();) {
            final Object cur = iter.next ();
            assertTrue (from.contains (cur));
            from.remove (cur);
        }
    }

    public void testIteratorBefore () {
        if (_store.preservesOrdering ())
            doTestIteratorWithIndex (_store.before ());
        else {
            checkEquals ("acabd", _store);

            Iterator iter = _store.before ();
            assertTrue (iter.hasNext ());
            assertFalse (iter.hasCurrent ());

            final java.util.List ref = new ArrayList (Arrays.asList (_store.toArray ()));
            while (iter.hasNext ()) {
                iter.next ();
                assertTrue (iter.hasCurrent ());
                assertTrue (ref.contains (iter.current ()));
                ref.remove (iter.current ());
            }
            assertTrue (ref.isEmpty ());

            assertFalse (iter.hasNext ());
            assertTrue (iter.hasCurrent ());

            for (int i = 1; i <= 5; i++) {
                setUp ();
                final java.util.List ref2 = new ArrayList (Arrays.asList (_store.toArray ()));

                iter = _store.before ();
                iter.next (i);
                final Object toBeRemoved = iter.current ();
                iter.remove ();
                removeSingly (ref2, Arrays.asList (_store.toArray ()));

                assertEquals (1, ref2.size ());
                assertEquals (toBeRemoved, ref2.get (0));
            }
        }
    }

    public void testIteratorBegin () {
        if (_store.preservesOrdering ()) {
            final Iterator iter = _store.begin ();
            checkIter (_store, iter, 0, "a");

            iter.previousIndex ();
            doTestIteratorWithIndex (_store.before ());
        }
        else {
            checkEquals ("acabd", _store);

            Iterator iter = _store.begin ();
            assertTrue (iter.hasNext ());
            assertTrue (iter.hasCurrent ());

            final java.util.List ref = new ArrayList (Arrays.asList (_store.toArray ()));
            while (iter.hasCurrent ()) {
                assertTrue (iter.hasCurrent ());
                assertTrue (ref.contains (iter.current ()));
                ref.remove (iter.current ());
                iter.next ();
            }
            assertTrue (ref.isEmpty ());

            assertFalse (iter.hasNext ());
            assertFalse (iter.hasCurrent ());

            for (int i = 0; i <= 4; i++) {
                setUp ();
                final java.util.List ref2 = new ArrayList (Arrays.asList (_store.toArray ()));

                iter = _store.begin ();
                iter.next (i);
                final Object toBeRemoved = iter.current ();
                iter.remove ();
                removeSingly (ref2, Arrays.asList (_store.toArray ()));

                assertEquals (1, ref2.size ());
                assertEquals (toBeRemoved, ref2.get (0));
            }
        }
    }


    public void testIteratorEnd () {
        if (!_iterSupportsPrevious)
            return;

        final Iterator iter = _store.end ();
        checkIter (_store, iter, 4, "d");
        iter.previous (5);
        doTestIteratorWithIndex (iter);
    }

    public void testIteratorAfter () {
        if (!_iterSupportsPrevious)
            return;

        final Iterator iter = _store.after ();
        checkIter (_store, iter, 5, null);
        iter.previous (6);
        doTestIteratorWithIndex (iter);
    }

    public void testIteratorIter () {
        if (!_store.preservesOrdering ())
            return;

        final Iterator iter = _store.iter (3);
        checkIter (_store, iter, 3, "b");
        iter.previous (4);
        doTestIteratorWithIndex (iter);
    }

    public void testIteratorFirst1 () {
        assertNull (_store.first ("X"));

        Iterator iter = _store.first ("a");
        checkIter (_store, iter, 0, "a");

        if (_iterSupportsPrevious) {
            iter.previous ();
            doTestIteratorWithIndex (iter);
        }
    }

    public void testIteratorFirst2 () {
        final Iterator iter = _store.first ("d");
        checkIter (_store, iter, 4, "d");

        if (_iterSupportsPrevious) {
            iter.previous (5);
            doTestIteratorWithIndex (iter);
        }
    }

    public void testIteratorLast1 () {
        if (!_iterSupportsPrevious)
            return;

        assertNull (_store.last ("X"));
        Iterator iter = _store.last ("a");
        checkIter (_store, iter, 2, "a");
        iter.previous (3);
        doTestIteratorWithIndex (iter);
    }

    public void testIteratorLast2 () {
        if (!_iterSupportsPrevious)
            return;

        final Iterator iter = _store.last ("c");
        checkIter (_store, iter, 1, "c");
        iter.previous (2);
        doTestIteratorWithIndex (iter);
    }


    void doTestIteratorWithIndex (Iterator iter) {
        checkEquals ("acabd", _store);
        checkIter (_store, iter, -1, null);

        assertEquals ("a", iter.next ());
        checkIter (_store, iter, 0, "a");

        assertEquals ("b", iter.next (3));
        checkIter (_store, iter, 3, "b");

        assertEquals ("d", iter.next ());
        checkIter (_store, iter, 4, "d");

        iter.next ();
        checkIter (_store, iter, 5, null);
        iter.previous ();

        assertEquals ("a", iter.previous (2));
        checkIter (_store, iter, 2, "a");

        assertEquals ("c", iter.previous ());
        checkIter (_store, iter, 1, "c");

        assertEquals ("a", iter.previous ());
        checkIter (_store, iter, 0, "a");

        iter.previous ();
        checkIter (_store, iter, -1, null);
        iter.next ();


        iter.add ("A");
        checkEquals ("Aacabd", _store);
        checkIter (_store, iter, 1, "a");

        iter.next (2);
        checkIter (_store, iter, 3, "a");
        assertTrue (iter.replace ("X"));
        checkEquals ("AacXbd", _store);
        checkIter (_store, iter, 3, "X");

        checkEquals ("AacXbd", _store);
        iter.add ("Y");
        checkEquals ("AacYXbd", _store);
        checkIter (_store, iter, 4, "X");

        iter.previous ();
        iter.previous ();
        iter.remove ();
        checkIter (_store, iter, 2, "Y");

        iter.previous (2);
        iter.remove ();
        checkIter (_store, iter, 0, "a");

        iter.next (4);
        iter.remove ();
        checkIter (_store, iter, 4, null);
    }

    public void testAddRemove () {
        final Store store = create (new EqualsEqualityStrategy ());

        assertTrue (store.isEmpty ());
        assertEquals (0, store.size ());
        assertFalse (store.contains ("a"));

        store.append ("a");
        store.append ("b");
        store.prepend ("c");
        store.append ("d");
        store.prepend ("a");

        assertFalse (store.isEmpty ());
        assertEquals (5, store.size ());
        assertTrue (store.contains ("a"));
        assertTrue (store.contains ("b"));
        assertTrue (store.contains ("c"));
        assertTrue (store.contains ("d"));
        assertFalse (store.contains ("e"));

        if (_store.preservesOrdering ()) {
            assertEquals ("a", store.get (0));
            assertEquals ("c", store.get (1));
            assertEquals ("d", store.get (4));

            store.set (0, "A");
            store.set (1, "C");
            store.set (4, "D");

            assertEquals ("A", store.get (0));
            assertEquals ("C", store.get (1));
            assertEquals ("D", store.get (4));

            assertEquals ("A", store.remove (0));
            checkEquals ("CabD", store);
            assertEquals ("D", store.remove (3));
            checkEquals ("Cab", store);
            assertEquals ("a", store.remove (1));
            checkEquals ("Cb", store);

            assertEquals (2, store.size ());
            assertFalse (store.isEmpty ());

            store.clear ();
            assertEquals (0, store.size ());
            assertTrue (store.isEmpty ());

            store.prepend ("a");
            store.clear ();

            store.append ("a");
            store.append ("b");
            store.prepend ("c");
            store.append ("d");
            store.prepend ("a");

            checkEquals ("acabd", store);
        }
    }
}

//todo: Exceptions